/***************************************************************************
 *
 * Copyright (C) 2001 International Business Machines
 * All rights reserved.
 *
 * This file is part of the GPFS mmfslinux kernel module.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer. 
 *  2. Redistributions in binary form must reproduce the above copyright 
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution. 
 *  3. The name of International Business Machines may not be used to endorse 
 *     or promote products derived from this software without specific prior 
 *     written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR 
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Alternatively, provided that this notice is retained in full, this
 * software may be distributed under the terms of the GNU General
 * Public License ("GPL") as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any later version,
 * in which case the provisions of the GPL apply INSTEAD OF those given above.
 *
 ****************************************************************************/
/* @(#)64       1.107.1.6  src/avs/fs/mmfs/ts/kernext/gpl-linux/cxiCache.c, mmfs, avs_rhin, rhins015a 6/22/12 10:21:55 */
/*
 * Cache (NFS) operations
 *
 * Contents:
 *   cxiCacheInit
 *   cxiCacheExit
 *   cxiCacheLookupRoot
 *   cxiCacheLookup
 *   cxiCacheOpen
 *   cxiCacheClose
 *   cxiCacheCreate
 *   cxiCacheRemove
 *   cxiCacheLink
 *   cxiCacheSymlink
 *   cxiCacheGetattr
 *   cxiCacheSetattr
 *   cxiCacheReaddir
 *   cxiCacheRead
 *   cxiCacheWrite
 *   cxiCacheRename
 *   cxiCacheReadlink
 *   cxiCacheNfsOpen
 *   cxiCacheGetFilename
 */

#include <Shark-gpl.h>
#include <Trace.h>
#include <linux2gpfs.h>

#include <linux/nfs_fs.h>
#include <linux/file.h>
#include <linux/fadvise.h>
/* #include <linux/fsnotify.h> */

#if LINUX_KERNEL_VERSION < 2063000
/* Mapping of S_IF* to NFS types */
static UInt32 cache_ftypes[] = 
{
        NFNON,  NFFIFO, NFCHR, NFNON,
        NFDIR,  NFNON,  NFBLK, NFNON,
        NFREG,  NFNON,  NFLNK, NFNON,
        NFSOCK, NFNON,  NFLNK, NFNON,
};
#endif

/* NFS function to get a Linux inode from the filehandle */
static struct inode *
(*fhget)(struct super_block *, struct nfs_fh *, struct nfs_fattr *) = NULL;

/* Copy NFS filehandle; identical to nfs_copy_fh which isn't present on all kernels */
static inline void
cache_copy_fh(struct nfs_fh *targetP, const struct nfs_fh *sourceP)
{
  targetP->size = sourceP->size;
  memcpy(targetP->data, sourceP->data, sourceP->size);
}

/* Compare two NFS filehandles */
static inline int
cache_compare_fh(const struct nfs_fh *a, const struct nfs_fh *b)
{
  return a->size != b->size || cxiMemcmp(a->data, b->data, a->size) != 0;
}

/* Convert pcache attr to NFS fattr */
static void
pattr2fattr(struct super_block *sbP, pcacheAttr_t *pattrP, struct nfs_fattr *fattrP)
{
  u32 time[2];

#if LINUX_KERNEL_VERSION < 2063000
  fattrP->type = cache_ftypes[(pattrP->pa_mode & S_IFMT) >> 12];
  DBGASSERT(fattrP->type != NFNON);
#endif
  fattrP->valid = NFS_ATTR_FATTR | NFS_ATTR_FATTR_V3;
  if (NFS_PROTO(sbP->s_root->d_inode)->version == 4)
    fattrP->valid |= NFS_ATTR_FATTR_V4;
  fattrP->nlink = pattrP->pa_nlink;
  /* FIX: There are cases where we are called for a file that's locally deleted
     (for example to replay a remove). In these cases nlink will be 0 but
     NFS will complain, so we set it to 1 */
  if (fattrP->nlink == 0)
  {
    fattrP->nlink = 1;
  }
  fattrP->fileid = (u64)pattrP->pa_ino;
  fattrP->mode = pattrP->pa_mode;
  fattrP->size = pattrP->pa_size;
#ifdef HAS_NFS_SUBMOUNTS
  /* We don't store server-provided fsid in GPFS, so use the one from the NFS sb */
  memcpy(&NFS_FATTR_FSID(fattrP), &NFS_SB(sbP)->fsid, sizeof(NFS_SB(sbP)->fsid));
  fattrP->time_start = 0;
#else
  NFS_FATTR_FSID(fattrP).major = NFS_FATTR_FSID(fattrP).minor = 0; /* unused */
#endif
  fattrP->rdev = ((pattrP->pa_rdev>>16)<<20)|(pattrP->pa_rdev & 0xFFFF);
  //CXITIME_TO_INODETIME(pattrP->pa_atime, fattrP->atime);
  CXITIME_TO_INODETIME(pattrP->pa_mtime, fattrP->mtime);
  CXITIME_TO_INODETIME(pattrP->pa_ctime, fattrP->ctime);
  time[0] = (u32)pattrP->pa_ctime.tv_sec;
  time[1] = (u32)pattrP->pa_ctime.tv_nsec;
  memcpy(&fattrP->change_attr, &time, sizeof(u64));
  fattrP->uid = pattrP->pa_uid;
  fattrP->gid = pattrP->pa_gid;
  //fattrP->du.nfs3.used = (u64)pattrP->pa_blocks << 9;
}

/* Fill up NFS file attributes from the inode */
static void inode2pattr(struct inode *inodeP, pcacheAttr_t *pattrP)
{
  struct nfs_inode *nfsiP = NFS_I(inodeP);
  cxiFH_t *fhP = &pattrP->pa_fh;
  __u32 *fh = (__u32 *)fhP->data;

  pattrP->pa_mode = inodeP->i_mode;
  pattrP->pa_uid = inodeP->i_uid;
  pattrP->pa_gid = inodeP->i_gid;
  pattrP->pa_rdev = cxiDevToDev32(inodeP->i_rdev);
  pattrP->pa_ino = (cxiInoSys_t)nfsiP->fileid;
  pattrP->pa_nlink = inodeP->i_nlink;
  pattrP->pa_size = inodeP->i_size;
  CXITIME_FROM_INODETIME(pattrP->pa_mtime, inodeP->i_mtime);
  CXITIME_FROM_INODETIME(pattrP->pa_ctime, inodeP->i_ctime);
  CXITIME_FROM_INODETIME(pattrP->pa_atime, inodeP->i_atime);
  cache_copy_fh((void *)fhP, NFS_FH(inodeP));
  TRACE11(TRACE_PCACHE, 9, TRCID_CACHE_INODE2PATTR,
          "inode2pattr: inode 0x%X mode %o uid %d gid %d rdev %d inode %lld "
          "size %d mtime %ld ctime %ld atime.nsec %lx fhsize %d",
          inodeP, inodeP->i_mode, inodeP->i_uid, inodeP->i_gid, inodeP->i_rdev,
          nfsiP->fileid, inodeP->i_size, inodeP->i_mtime.tv_sec,
          inodeP->i_ctime.tv_sec, pattrP->pa_atime.tv_nsec, fhP->size);
  TRACE10(TRACE_PCACHE, 9, TRCID_CACHE_INODE2PATTR_2,
         "inode2pattr: fh: size %u %08x %08x %08x %08x %08x %08x %08x %08x %08x",
         fhP->size, fh[0], fh[1], fh[2], fh[3], fh[4], fh[5], fh[6], fh[7], fh[8]);
}

/* Revalidate a dentry that is found in the cache. Depending on the
   intent specified in the nameidata, this may cause the file to be
   opened at the server */
static inline struct dentry *
cache_revalidate(struct dentry *dP, struct nameidata *ndP)
{
  int rc = 0;

  if (dP->d_op && dP->d_op->d_revalidate)
  {
    rc = dP->d_op->d_revalidate(dP, ndP);
    if (rc > 0)
      goto out;
    if (rc == 0)
    {
      if (!d_invalidate(dP))
      {
        dput(dP);
        dP = NULL;
      }
    }
    else
    {
      dput(dP);
      dP = ERR_PTR(rc);
    }
  }
 out:
  TRACE2(TRACE_PCACHE, 4, TRCID_CACHE_REVALIDATE,
         "cache_revalidate exit: dP 0x%lX rc %d",
         dP, rc);
  return dP;
}

/* Lookup a dentry given its name and the inode of the parent directory.
   Returns with a hold on the dentry.  This is similar to Linux's generic
   do_lookup() but that is not exported for use by external modules */
static struct dentry *
cache_lookup(struct vfsmount *vfsP, struct dentry *ddP, const char *nameP, struct nameidata *ndP, int create)
{
  struct dentry *dP = NULL;
  struct inode *dirP = ddP->d_inode;
  struct file *fileP = NULL;
  struct qstr name;
  int code = 0;
#if defined(USE_ALLOC_FILE)
  struct path path;
#endif

  /* If the name starts with a '/', we are probably looking up '.' */
  while (*nameP == '/')
    nameP++;

  if (!*nameP || (*nameP == '.' && !*(nameP+1)))
  {
    dP = cache_revalidate(dget(ddP), NULL);
    code = 1;
    if (dP == NULL)
    {
      /* If we fail to revalidate the root, return ESTALE */
      dP = ERR_PTR(-ESTALE);
      code = 11;
    }
    goto out;
  }

  name.name = nameP;
  name.len = strlen(nameP);
  name.hash = full_name_hash(name.name, name.len);

  INODE_LOCK(dirP);

  /* Check if the dentry exists in the dcache */
  dP = d_lookup(ddP, &name);
  if (dP != NULL)
  {
    INODE_UNLOCK(dirP);
    TRACE2(TRACE_PCACHE, 4, TRCID_CACHE_LOOKUP_1,
          "cache_lookup: found cached dentry 0x%lX for %s",
          dP, nameP);
    /* NFS only supports OPEN on regular files */
    if (dP->d_inode != NULL &&
        S_ISREG(dP->d_inode->i_mode) &&
        (ndP->flags & LOOKUP_OPEN))
    {
#ifdef LOOKUP_INTENT
#if defined(USE_ALLOC_FILE)
      path.mnt = vfsP;
      path.dentry = dP;
      /* D.781282 - Simulate get_empty_filp() behaviour via alloc_file() */
      fileP = alloc_file(&path, INTENT_OPEN_FLAGS(ndP->intent), NULL);
      if (fileP)
      {
        fileP->f_dentry   = NULL;
        fileP->f_mapping  = NULL;
        fileP->f_mode     = 0;
        fileP->f_op       = NULL;
      }
      else
#else
      fileP = get_empty_filp();
      if (fileP == NULL)
#endif // USE_ALLOC_FILE
      {
        dput(dP);
        dP = ERR_PTR(-ENFILE); /* FIX: ENOENT instead? */
        code = 2;
        goto out;
      }
      ndP->intent.open.file = fileP;
#endif
      dP = cache_revalidate(dP, ndP);
    }
    else
      dP = cache_revalidate(dP, NULL);
    if (dP == NULL)
      INODE_LOCK(dirP);
  }

  /* Check again for dP since revalidate may have invalidated it */
  if (dP == NULL)
  {
    struct dentry *newP = NULL;

    newP = d_alloc(ddP, &name);
    if (newP == NULL)
    {
      dP = ERR_PTR(-ENOMEM);
      code = 4;
      goto out;
    }
#ifdef LOOKUP_INTENT
    /* Allocate a file struct for open */
    if (ndP->flags & LOOKUP_OPEN)
    {
#if defined(USE_ALLOC_FILE)
      path.mnt = vfsP;
      /* D.781282 - Simulate get_empty_filp(); Pretend with parent dentry */
      path.dentry = ddP;
      fileP = alloc_file(&path, INTENT_OPEN_FLAGS(ndP->intent), NULL);
      if (fileP)
      {
        fileP->f_dentry   = NULL;
        fileP->f_mapping  = NULL;
        fileP->f_mode     = 0;
        fileP->f_op       = NULL;
      }
      else
#else
      fileP = get_empty_filp();
      if (fileP == NULL)
#endif
      {
        dP = ERR_PTR(-ENFILE);  /* FIX: ENOENT instead? */
        code = 3;
        goto out;
      }
      ndP->intent.open.file = fileP;
    }
#endif

    DBGASSERT(dirP->i_op->lookup != NULL);
    dP = dirP->i_op->lookup(dirP, newP, ndP);
    if (dP)
    {
      dput(newP);
#if (LINUX_KERNEL_VERSION >= 2061800)
      if (!IS_ERR(dP) && dP->d_inode != NULL && S_ISDIR(dP->d_inode->i_mode))
      {
        /* If the lookup returns an "anonymous" (no-name) dentry, we need
           to re-fill it with the name being looked up. This happens for
           directory dentries that are created from attribute lookup using
           d_alloc_anon */
        struct dentry *resP;

        /* We need to allocate a new dentry with the name being looked up,
           because i_op->lookup trashed it */
        newP = d_alloc(ddP, &name);
        if (newP == NULL)
        {
          dput(dP);
          dP = ERR_PTR(-ENOMEM);
          code = 6;
          goto out;
        }
        resP = d_materialise_unique(newP, dP->d_inode); 
        if (resP)
          dput(newP);
        else
          resP = newP;
        dput(dP);
        dP = resP;
      }
#endif
    }
    else
      dP = newP;
    INODE_UNLOCK(dirP);
  }

  /* Is this a negative dentry? */
  if (!create && !IS_ERR(dP) && dP->d_inode == NULL)
  {
    dput(dP);
    dP = ERR_PTR(-ENOENT);
    code = 5;
  }
 out:
  TRACE5(TRACE_PCACHE, 4, TRCID_CACHE_LOOKUP,
         "cache_lookup exit: ddP 0x%lX name %s flags 0x%X ret dP 0x%lX code %d",
         ddP, nameP, ndP->flags, dP, code);

  if (fileP) // USE_ALLOC_FILE
    TRACE4(TRACE_PCACHE, 6, TRCID_CACHE_LOOKUP_EXIT1,
           "cache_lookup exit: fileP 0x%lX f_dentry 0x%lX mode %d open %d",
           fileP, fileP->f_dentry, fileP->f_mode, ndP->flags & LOOKUP_OPEN ? 1 : 0);

  return dP;
}

/**
 * @dentry: the directory in which to find a name
 * @name:   a pointer to a %NAME_MAX+1 char buffer to store the name
 * @child:  the dentry for the child directory.
 *
 * calls readdir on the parent until it finds an entry with
 * the same inode number as the child, and returns that.
 */
struct getdents_callback
{
  char *nameP;
  unsigned long ino;
  int found;
  int sequence;
};

static int filldir_one(void * __bufP, const char * nameP, int len,
                       loff_t pos, u64 ino, unsigned int d_type)
{
  struct getdents_callback *bufP = __bufP;
  int result = 0;

  bufP->sequence++;
  if (bufP->ino == ino) 
  {
    memcpy(bufP->nameP, nameP, len);
    bufP->nameP[len] = '\0';
    bufP->found = 1;
    result = -1;
  }
  return result;
}

/* Do a readdir on the parent to find the matching name */
static int cache_get_name(struct vfsmount *mntP, struct dentry *dentryP,
                          char *nameP, struct dentry *childP)
{
  struct inode *dirP = dentryP->d_inode;
  int error = 0;
  struct file *fileP = NULL;
  struct getdents_callback buffer;

  error = -ENOTDIR;
  if (!dirP || !S_ISDIR(dirP->i_mode))
    goto out;
  error = -EINVAL;
  DBGASSERT(dirP->i_fop != NULL);

  /* Open the directory ... */
  fileP = DENTRY_OPEN(dget(dentryP), mntP, O_RDONLY);
  if (IS_ERR(fileP))
  {
    error = PTR_ERR(fileP);
    goto out;
  }
  DBGASSERT(fileP->f_op->readdir != NULL);

  buffer.nameP = nameP;
  buffer.ino = childP->d_inode->i_ino;
  buffer.found = 0;
  buffer.sequence = 0;
  while (1) 
  {
    int old_seq = buffer.sequence;
    error = vfs_readdir(fileP, (filldir_t)filldir_one, &buffer);
    if (error < 0)
      break;

    error = 0;
    if (buffer.found)
      break;
    error = -ENOENT;
    if (old_seq == buffer.sequence)
      break;
  }
  fput(fileP);

out:
  return error;
}

/* Find a dentry associated with the specified inode.
   Returns with a hold on the dentry
   - if parentP is given, use it to find the correct one
   - if nameP is given, use it to find the matching name
   On error, releases hold on inode */
static struct dentry *
cache_find_dentry(struct vfsmount *mntP, struct dentry *parentP,
                  struct inode *inodeP, const char *nameP)
{
  struct dentry *resP = NULL;
  struct dentry *dirP = NULL;
  int noprogress = 0;
  int err = 0;

  /* If name is specified, try to find the right dentry */
  if (nameP != NULL)
  {
    struct dentry *ldP;
#ifdef DCACHE_LOCK_IS_GONE
    spin_lock(&inodeP->i_lock);
#else
    spin_lock(&dcache_lock);
#endif
    list_for_each_entry(ldP, &inodeP->i_dentry, d_alias)
    {
      /* Check if the name matches */
      if (nameP && cxiMemcmp(ldP->d_name.name, nameP, ldP->d_name.len) == 0)
      {
#ifdef DCACHE_LOCK_IS_GONE
        resP = dget(ldP);
#else
        resP = dget_locked(ldP);
#endif
        break;
      }
    }
#ifdef DCACHE_LOCK_IS_GONE
    spin_unlock(&inodeP->i_lock);
#else
    spin_unlock(&dcache_lock);
#endif
    if (resP == NULL) /* didn't find it */
    {
      if (parentP != NULL)
      {
        struct qstr q;

        q.name = nameP;
        q.len = strlen(nameP);
        q.hash = full_name_hash(q.name, q.len);
        resP = d_alloc(parentP, &q);
        if (resP != NULL)
          d_add(resP, inodeP);
      }
      else
        resP = d_alloc_anon(inodeP);
    }
  }
  else
    resP = d_alloc_anon(inodeP);

  /* Return error if we couldn't get a valid dentry;
     need to release hold on inode */
  if (resP == NULL)
  {
    resP = ERR_PTR(-ENOENT);
    goto out_iput;
  }
  else if (IS_ERR(resP))
    goto out;

  if (S_ISDIR(resP->d_inode->i_mode) && (resP->d_flags & DCACHE_DISCONNECTED))
  {
    /* the directory is not connected, we must connect it */
    ;
  }
  else
    goto out;

  dirP = dget(resP);

  /* Note: The following code is similar in concept to find_exported_dentry
     used by exportfs */

  /* Now we need to make sure that dirP is properly connected.
   * It may already be, as the flag isn't always updated when connection
   * happens.
   * So, we walk up parent links until we find a connected directory,
   * or we run out of directories.  Then we find the parent, find
   * the name of the child in that parent, and do a lookup.
   * This should connect the child into the parent
   * We then repeat.
   */
  while ((dirP->d_flags & DCACHE_DISCONNECTED) && noprogress++ < 10)
  {
    struct dentry *pdP = dirP;

    dget(pdP);
    spin_lock(&pdP->d_lock);
    while (!IS_ROOT(pdP) && (pdP->d_parent->d_flags & DCACHE_DISCONNECTED))
    {
      struct dentry *pddP = pdP->d_parent;
      dget(pddP);
      spin_unlock(&pdP->d_lock);
      dput(pdP);
      pdP = pddP;
      spin_lock(&pdP->d_lock);
    }
    spin_unlock(&pdP->d_lock);

    if (!IS_ROOT(pdP))
    {
      /* found a connected parent */
      spin_lock(&pdP->d_lock);
      pdP->d_flags &= ~DCACHE_DISCONNECTED;
      spin_unlock(&pdP->d_lock);
      noprogress = 0;
    }
    else
    {
      /* Note: This code is intentionally duplicated pending the fix below */
      /* FIX: We really need to check here if we are at the root of the NFS
       * mount with something like (pdP == mntP->mnt_sb->s_root) but with
       * kernels > 2.6.18, the NFS superblock is shared by all mounts to a server
       * so the condition may not work
       */
      spin_lock(&pdP->d_lock);
      pdP->d_flags &= ~DCACHE_DISCONNECTED;
      spin_unlock(&pdP->d_lock);
      noprogress = 0;
    }
    dput(pdP);
  }

  if (dirP->d_flags & DCACHE_DISCONNECTED)
  {
    /* something went wrong - oh-well */
    dput(resP);
    resP = ERR_PTR(-ESTALE);
    goto out_dput;
  }

  /* if we weren't after a directory, have one more step to go */
  if (resP != dirP)
  {
    char *buf;
    struct dentry *nresP;

    buf = (char *) __get_free_page(GFP_USER);
    if (buf == NULL)
    {
      dput(resP);
      resP = ERR_PTR(-ENOMEM);
      goto out_dput;
    }

    err = cache_get_name(mntP, dirP, buf, resP);
    if (!err)
    {
      INODE_LOCK(dirP->d_inode);
      nresP = lookup_one_len(buf, dirP, strlen(buf));
      INODE_UNLOCK(dirP->d_inode);
      if (!IS_ERR(nresP))
      {
        if (nresP->d_inode)
        {
          dput(resP);
          resP = nresP;
        }
        else
          dput(nresP);
      }
    }
    free_page((unsigned long) buf);
  }

 out_dput:
  dput(dirP);
 out_iput:
  if (IS_ERR(resP))
    iput(inodeP);
 out:
  TRACE3(TRACE_PCACHE, 9, TRCID_CACHE_FIND_DENTRY_EXIT,
         "cache_find_dentry: inode %lld name %s ret dentry 0x%lX",
         inodeP->i_ino, nameP, resP);
  return resP;
}

/* Open a file using the dentry - a file struct may have been preallocated */
static struct file *
cache_dentry_open(struct dentry *dentryP, struct vfsmount *mntP, int flags, struct file *fileP)
{
  struct inode *inodeP;
  struct file *fP = fileP;
  int err;

  if (fP == NULL)
    return DENTRY_OPEN(dentryP, mntP, flags);

  /* Fill existing file */
  fP->f_flags = flags;
  fP->f_mode = ((flags+1) & O_ACCMODE) | FMODE_LSEEK |
               FMODE_PREAD | FMODE_PWRITE;
  inodeP = dentryP->d_inode;
  DBGASSERT(inodeP != NULL);
  if (fP->f_mode & FMODE_WRITE)
  {
    err = get_write_access(inodeP);
    if (err)
    {
      fput(fP);
      return ERR_PTR(err);
    }
  }
  fP->f_mapping = inodeP->i_mapping;
  fP->f_dentry = dentryP;
  fP->f_vfsmnt = mntP;
  fP->f_pos = 0;
  fP->f_op = fops_get(inodeP->i_fop);
  err = fP->f_op->open(inodeP, fP);
  if (err == 0)
  {
    fP->f_flags &= ~(O_CREAT | O_EXCL | O_NOCTTY | O_TRUNC);
    /* file_ra_state_init(&fP->f_ra, fP->f_mapping->host->i_mapping); */
    return fP;
  }

  /* Error */
  fops_put(fP->f_op);
  if (fP->f_mode & FMODE_WRITE)
    put_write_access(inodeP);
  fP->f_op = NULL;
  fput(fP);
  return ERR_PTR(err);
}

struct nfs_find_desc
{
  cxiIno64_t ino;
  struct nfs_fh *fhP;
};
  
static int
nfs_find_inode(struct inode *inodeP, void *opaqueP)
{
  struct nfs_find_desc *descP = (struct nfs_find_desc *)opaqueP;
  struct nfs_fh *fhP = descP->fhP;
  cxiIno64_t ino = descP->ino;
  
  if (inodeP->i_ino != ino ||
      cache_compare_fh(NFS_FH(inodeP), fhP) ||
      is_bad_inode(inodeP) ||
      NFS_STALE(inodeP))
    return 0;
  return 1;
}
  
/* Get NFS inode from filehandle and attributes
   Returns with a hold on the inode */
static struct inode *
nfs_iget(void *vfsP, pcacheAttr_t *pattrP, int getNew)
{
  struct vfsmount *mntP = (struct vfsmount *)vfsP;
  struct nfs_fh *fhP = (struct nfs_fh *)&pattrP->pa_fh;
  struct nfs_fattr fattr;
  struct inode *inodeP = ERR_PTR(-ENOENT);
  __u32 *fh = (__u32 *)fhP->data;
  struct nfs_find_desc desc = 
  {
    .fhP = fhP,
    .ino = pattrP->pa_ino
  };

  if (fhP->size == 0)
    goto out;

  /* We don't want to call fhget() directly here before checking if the inode
     is in the cache. If we do, a cached inode with (valid) attributes will
     be updated with the ones we are passing to fhget(). We only want to use
     the local cached attributes to help construct a new inode. So we check first
     if there is a cached inode. If we find one, we return it directly - it will
     be revalidated later */
  inodeP = ilookup5(mntP->mnt_sb, desc.ino, nfs_find_inode, &desc);
  if (inodeP == NULL && getNew)
  {
    pattr2fattr(mntP->mnt_sb, pattrP, &fattr);
    inodeP = fhget(mntP->mnt_sb, fhP, &fattr);
  }

 out:
  TRACE3(TRACE_PCACHE, 4, TRCID_NFS_IGET,
         "nfs_iget: inode %lld, fh size %d, inodeP 0x%lX",
         pattrP->pa_ino, fhP->size, inodeP);
  TRACE10(TRACE_PCACHE, 9, TRCID_NFS_IGET_FH,
         "nfs_iget: fh: size %u %08x %08x %08x %08x %08x %08x %08x %08x %08x",
         fhP->size, fh[0], fh[1], fh[2], fh[3], fh[4], fh[5], fh[6], fh[7], fh[8]);
  return inodeP;
}

/* Get NFS dentry from filehandle and attributes.
   name (and parent dentry) is used to find the matching dentry
   Returns with a hold on the dentry */
static struct dentry *
nfs_dget_name(void *vfsP, struct dentry *parentP,
              const char *nameP, pcacheAttr_t *pattrP)
{
  struct inode *inodeP = NULL;
  struct dentry *dentryP = ERR_PTR(-ENOENT);

  inodeP = nfs_iget(vfsP, pattrP, true);
  if (!IS_ERR(inodeP))
  {
    dentryP = cache_find_dentry(vfsP, parentP, inodeP, nameP);
    if (!IS_ERR(dentryP))
    {
      struct kstat stat;
      int rc = 0;
      /* We need to refresh the inode since we used a (possibly stale) fh
         to construct the dentry from it */
      rc = vfs_getattr(vfsP, dentryP, &stat);
      TRACE3(TRACE_PCACHE, 4, TRCID_NFS_GETATTR,
             "nfs_dget_name: vfs_getattr inode 0x%lX name %s ret %d",
             inodeP, dentryP->d_name.name, rc);
      if (rc != 0)
      {
        dput(dentryP);
        dentryP = ERR_PTR(rc);
      }
    }
  }
  else
    dentryP = (struct dentry *)inodeP; /* err */

  return dentryP;
}

/* Get NFS dentry from filehandle and attributes.
   Returns with a hold on the dentry */
static inline struct dentry *
nfs_dget(void *vfsP, pcacheAttr_t *pattrP)
{
  return nfs_dget_name(vfsP, NULL, NULL, pattrP);
}

/*
 * Cache operations
 */

/* Initialization routine called at mount time to establish link
   between a fileset and NFS vfsmount. Returns with a hold on
   the NFS mount so it doesn't disappear under us.*/
int cxiCacheInit(const char *mntPathP, void *addrP, void **vfsPP)
{
  struct nameidata nd;
  char *fsname;
  int rc = -ENODEV, code = 0;

  rc = PATH_LOOKUP(mntPathP, LOOKUP_FOLLOW | LOOKUP_DIRECTORY, &nd);
  if (rc != 0)
  {
    code = 1;
    goto out;
  }

  /* Make sure the mount path is for NFS */
  fsname = (char *)ND_PATH(nd).dentry->d_sb->s_type->name;
  if (strcmp(fsname, "nfs") != 0 && strcmp(fsname, "nfs4") != 0)
  {
    code = 2;
    rc = -ENODEV;
    goto out_put;
  }

  if (addrP != NULL)
  {
    /* Set the address of an unexported function - nfs_fhget */
    fhget = addrP;
  }

  /* Put a hold count on the vfsmount */
  *vfsPP = mntget(ND_PATH(nd).mnt);
 
  rc = 0;

out_put:
  PATH_PUT(&nd);

out:
  TRACE4(TRACE_PCACHE, 4, TRCID_CXICACHE_INIT,
         "cxiCacheInit: pathP %s remote vfsP 0x%lX rc %d code %d",
         mntPathP, *vfsPP, rc, code);
  return -rc;
}

/* Exit routine called at umount time to release the hold on
   the NFS mount established at GPFS mount time. Also closes the special file */
int cxiCacheExit(void *vfsP, void *fileP)
{
  DBGASSERT(vfsP != NULL);

  if (fileP != NULL)
    cxiCacheClose(fileP, NULL);
  mntput(vfsP);
  TRACE2(TRACE_PCACHE, 4, TRCID_CXICACHE_EXIT,
         "cxiCacheExit: nfs vfsP 0x%lX xattr fileP 0x%lX", vfsP, fileP);
  return 0;
}

/* Return NFS fh and attrs of root */
int
cxiCacheLookupRoot(void *vfsP, pcacheAttr_t *pattrP)
{
  struct dentry *rootP = ((struct vfsmount *)vfsP)->mnt_root;
  int rc = -ENOENT;

  if (rootP && rootP->d_inode)
  {
    inode2pattr(rootP->d_inode, pattrP);
    rc = 0;
  }
  TRACE2(TRACE_PCACHE, 4, TRCID_CXICACHE_LOOKUP_ROOT,
         "cxiCacheLookupRoot: returned ino %lld rc %d",
         pattrP->pa_ino, rc);
  return -rc;
}

/* Do lookup on a name, given the parent fh and attrs,
   and if successful, return resulting fh and attrs */
int
cxiCacheLookup(void *vfsP, pcacheAttr_t *pattrP, char *nameP, pcacheAttr_t *cpattrP)
{
  struct dentry *ddP = NULL, *dP = NULL;
  struct nameidata nd = {};
  int rc = 0;

  /* Get dentry of parent */
  if (*nameP == '/')
    ddP = dget(((struct vfsmount *)vfsP)->mnt_root);
  else
    ddP = nfs_dget(vfsP, pattrP);
  if (IS_ERR(ddP))
  {
    rc = PTR_ERR(ddP);
    goto out;
  }
  DBGASSERTRC(S_ISDIR(ddP->d_inode->i_mode), ddP->d_inode->i_ino, 0, 0);
  ND_PATH(nd).mnt = mntget(vfsP);
  ND_PATH(nd).dentry = ddP;  /* We already have a hold on ddP */

  dP = cache_lookup(vfsP, ddP, nameP, &nd, 0);
  if (IS_ERR(dP))
  {
    rc = PTR_ERR(dP);
    goto out_dput;
  }

  inode2pattr(dP->d_inode, cpattrP);
  dput(dP);

 out_dput:
  PATH_PUT(&nd);
 out:
  TRACE4(TRACE_PCACHE, 4, TRCID_CXICACHE_LOOKUP,
         "cxiCacheLookup: name %s, dir %lld, returned ino %lld rc %d",
          nameP, (pattrP != NULL)?pattrP->pa_ino:-1LL, (cpattrP != NULL)?cpattrP->pa_ino:-1LL, rc);
  return -rc;
}

/* Open a file identified by name, fh and attrs (including parent's) */
int
cxiCacheOpen(void *vfsP, pcacheAttr_t *parentPattrP, const char *nameP, int flags,
             int mode,  uid_t uid, gid_t gid,  
             pcacheAttr_t *pattrP, void **filePP, loff_t *fileSizeP)
{
  struct dentry *ddP = NULL, *dP = NULL;
  struct nameidata nd;
  struct file *fileP = NULL;
  char *fileNameP = NULL;
  int namei_flags = 0, lookup_flags = 0, rc = 0;

  TRACE2(TRACE_PCACHE, 4, TRCID_CXICACHE_OPEN_IN,
         "cxiCacheOpen: name %s flags 0%o", nameP, flags);

  /* Init return values */
  *filePP = NULL;
  *fileSizeP = 0;

  if (nameP == NULL || *nameP == '\0')
  {
    dP = nfs_dget(vfsP, parentPattrP);
    if (IS_ERR(dP))
    {
      rc = PTR_ERR(dP);
      goto out;
    }
    ddP = dget(dP->d_parent);
    fileNameP = (char *)dP->d_name.name;
  }
  else
  {
    /* Lookup up the dentries from the filehandle/attributes */
    ddP = nfs_dget(vfsP, parentPattrP);
    if (IS_ERR(ddP))
    {
      rc = PTR_ERR(ddP);
      goto out;
    }
    fileNameP = (char *)nameP;
    DBGASSERTRC(S_ISDIR(ddP->d_inode->i_mode), ddP->d_inode->i_ino, 0, 0);
  }

  ND_PATH(nd).mnt = mntget(vfsP);
  ND_PATH(nd).dentry = ddP; /* We already have a hold on ddP */

  /* We always want to support large files */
  flags |= O_LARGEFILE;

  /* Map open flags to open intent flags */
  namei_flags = flags;
  if ((namei_flags+1) & O_ACCMODE)
    namei_flags++;

  lookup_flags = LOOKUP_FOLLOW;
  if (flags & O_NOFOLLOW)
    lookup_flags &= ~LOOKUP_FOLLOW;
  if (flags & O_DIRECTORY)
    lookup_flags |= LOOKUP_DIRECTORY;

  nd.flags = lookup_flags | LOOKUP_OPEN;
  INTENT_OPEN_FLAGS(nd.intent) = namei_flags;
  INTENT_OPEN_CREATE_MODE(nd.intent) = mode;

#ifdef LOOKUP_INTENT
  /* Note: file struct below will be allocated by cache_lookup, because
     we do not want to do a open revalidate for directories */
  nd.intent.open.file = NULL;
#endif

  if (dP == NULL)
  {
    dP = cache_lookup(vfsP, ddP, fileNameP, &nd, (flags & O_CREAT));
    if (IS_ERR(dP))
    {
      rc = PTR_ERR(dP);
      goto out_release;
    }
    if ((dP->d_inode == NULL) && (flags & O_CREAT))
    {
      /* if we are here then O_CREAT flag is already set */
      rc = vfs_create(ddP->d_inode, dP, mode, &nd);
      if (rc != 0)
      {
        rc = -ENOENT;
        goto out_release; 
      }
      if (uid != 0 || gid != 0)
      {
        /* FIX: This check needs to be updated when we create as non-root user.
           If we run in the context of the caller, then no attribute update is required.
           Until then, if a non-root user performed the original create, we need
           to update remote uid/gid */
        struct iattr iattr;
        iattr.ia_valid = ATTR_UID|ATTR_GID;
        iattr.ia_uid = uid;
        iattr.ia_gid = gid;
        rc = dP->d_inode->i_op->setattr(dP, &iattr);
      }
    }
  }

  /* Initialize fileP if it hasn't already been done */
#ifdef LOOKUP_INTENT
  fileP = nd.intent.open.file;
#endif
  if (fileP == NULL || fileP->f_dentry == NULL)
  {
    fileP = cache_dentry_open(dget(dP), mntget(vfsP), flags, fileP);
    if (IS_ERR(fileP))
    {
      rc = PTR_ERR(fileP);
      goto out_dput;
    }
  }

#if LINUX_KERNEL_VERSION >= 2061800
  memset(&fileP->f_ra, 0, sizeof(fileP->f_ra));
  file_ra_state_init(&fileP->f_ra, fileP->f_mapping->host->i_mapping);
#endif

  *filePP = fileP;
  *fileSizeP = dP->d_inode->i_size;
  if (pattrP != NULL)
    inode2pattr(dP->d_inode, pattrP);

 out_dput:
  dput(dP);
 out_release:
  PATH_PUT(&nd);
 out:
  TRACE7(TRACE_PCACHE, 4, TRCID_CXICACHE_OPEN,
         "cxiCacheOpen: parent ino %lld name %s flags 0x%X ret ino %lld fileP 0x%lX size %lld rc %d",
         parentPattrP->pa_ino, (fileNameP != NULL) ? fileNameP : "NULL", flags,
         (rc == 0 && dP && dP->d_inode) ? dP->d_inode->i_ino : -1LL, fileP, *fileSizeP, rc);
  return -rc;
}

/* Create a new object and return its fh and attrs */
int
cxiCacheCreate(void *vfsP, pcacheAttr_t *parentPattrP, char *nameP, int mode,
               uid_t uid, gid_t gid, cxiDev_t rdev, pcacheAttr_t *cpattrP)
{
  struct dentry *ddP = NULL, *dP = NULL;
  struct qstr name;
  struct nameidata nd = {};
  dev_t dev;
  int umask;
  int code = 0, rc = 0;

  /* Get dentry of parent */
  ddP = nfs_dget(vfsP, parentPattrP);
  if (IS_ERR(ddP))
  {
    rc = PTR_ERR(ddP);
    code = 1;
    goto out;
  }
  if(!S_ISDIR(ddP->d_inode->i_mode))
  {
    rc = -ENOTDIR;
    code = 2;
    goto out;
  }

  name.name = nameP;
  name.len = strlen(nameP);
  name.hash = full_name_hash(name.name, name.len);

  ND_PATH(nd).mnt = mntget(vfsP);
  ND_PATH(nd).dentry = ddP; /* We already have a hold on ddP */

  dP = cache_lookup(vfsP, ddP, nameP, &nd, 1);
  if (IS_ERR(dP))
  {
    code = 3;
    rc = PTR_ERR(dP);
    goto out_put;
  }

  if (dP->d_inode != NULL)
  {
    if (cpattrP)
      inode2pattr(dP->d_inode, cpattrP);
    dput(dP);
    code = 4;
    rc = -EEXIST;
    goto out_put;
  }

  /* Use umask of 0 instead of process umask */
  umask = current->fs->umask;
  current->fs->umask = 0;
  switch (mode & S_IFMT)
  {
    case S_IFREG:
      /* Need to pass in a valid nameidata since NFS looks at the flags */
      nd.flags = LOOKUP_CREATE;
      INTENT_OPEN_FLAGS(nd.intent) = FMODE_READ | FMODE_WRITE | O_CREAT | O_EXCL;
      INTENT_OPEN_CREATE_MODE(nd.intent) = mode;
      rc = vfs_create(ddP->d_inode, dP, mode, &nd);
      break;

    case S_IFDIR:
      rc = VFS_MKDIR(vfsP, ddP->d_inode, dP, mode);
      break;

    case S_IFCHR:
    case S_IFBLK:
    case S_IFIFO:
    case S_IFSOCK:
      dev = ((rdev>>16)<<20)|(rdev & 0xFFFF);
      rc = VFS_MKNOD(vfsP, ddP->d_inode, dP, mode, dev);
      break;

    default:
      rc = -EINVAL;
  }
  /* Reset umask */
  current->fs->umask = umask;

  if (rc == 0)
  {
    if (uid != 0 || gid != 0)
    {
      /* FIX: This check needs to be updated when we create as non-root user.
         If we run in the context of the caller, then no attribute update is required.
         Until then, if a non-root user performed the original create, we need
         to update remote uid/gid */
      struct iattr iattr;
      iattr.ia_valid = ATTR_UID|ATTR_GID;
      iattr.ia_uid = uid;
      iattr.ia_gid = gid;
      rc = dP->d_inode->i_op->setattr(dP, &iattr);
    }
    if (cpattrP)
      inode2pattr(dP->d_inode, cpattrP);
  }
  dput(dP);

out_put:
  PATH_PUT(&nd);

out:
  TRACE7(TRACE_PCACHE, 4, TRCID_CXICACHE_CREATE,
         "cxiCacheCreate: name %s mode 0%o dir %ld uid %d gid %d rc %d code %d",
         nameP, mode, parentPattrP->pa_ino, uid, gid, rc, code);
  return -rc;
}

/* Remove an object, given its name and parent's fh and attrs */
int
cxiCacheRemove(void *vfsP, pcacheAttr_t *parentPattrP, char *nameP, int type)
{
  struct dentry *ddP = NULL;
  struct dentry *dP = NULL;
  struct nameidata nd = {};
  int code = 0, rc = 0;

  /* Get dentry of parent */
  ddP = nfs_dget(vfsP, parentPattrP);
  if (IS_ERR(ddP))
  {
    rc = PTR_ERR(ddP);
    code = 1;
    goto out;
  }
  DBGASSERTRC(S_ISDIR(ddP->d_inode->i_mode), ddP->d_inode->i_ino, 0, 0);

  ND_PATH(nd).mnt = mntget(vfsP);
  ND_PATH(nd).dentry = ddP; /* We already have a hold on ddP */

  dP = cache_lookup(vfsP, ddP, nameP, &nd, 0);
  if (IS_ERR(dP))
  {
    rc = PTR_ERR(dP);
    code = 2;
    goto out_dput;
  }

  switch (type)
  {
    case S_IFREG:
    case S_IFCHR:
    case S_IFBLK:
    case S_IFIFO:
    case S_IFSOCK:
      rc = VFS_UNLINK(vfsP, ddP->d_inode, dP);
      break;

    case S_IFDIR:
      rc = VFS_RMDIR(vfsP, ddP->d_inode, dP);
      break;

    default:
      code = 3;
      rc = -EINVAL;
  }
  dput(dP);

 out_dput:
  PATH_PUT(&nd);

 out:
  TRACE5(TRACE_PCACHE, 4, TRCID_CXICACHE_REMOVE,
         "cxiCacheRemove: name %s type 0x%X dir %lld, rc %d code %d",
         nameP, type, parentPattrP->pa_ino, rc, code);
  return -rc;
}

/* Create a hard link to an object identified by fh and attrs */
int
cxiCacheLink(void *vfsP, pcacheAttr_t *parentPattrP, pcacheAttr_t *pattrP, char *nameP)
{
  struct dentry *ddP = NULL, *sdP = NULL, *tdP = NULL;
  struct nameidata nd = {};
  int code = 0, rc = 0;

  /* Get dentry of parent and source */
  ddP = nfs_dget(vfsP, parentPattrP);
  if (IS_ERR(ddP))
  {
    rc = PTR_ERR(ddP);
    code = 1;
    goto out;
  }
  DBGASSERTRC(S_ISDIR(ddP->d_inode->i_mode), ddP->d_inode->i_ino, 0, 0);

  ND_PATH(nd).mnt = mntget(vfsP);
  ND_PATH(nd).dentry = ddP; /* We already have a hold on ddP */

  /* We don't have the source name, so can't call nfs_dget_name */
  sdP = nfs_dget(vfsP, pattrP);
  if (IS_ERR(sdP))
  {
    rc = PTR_ERR(sdP);
    code = 2;
    goto out_ddput;
  }

  tdP = cache_lookup(vfsP, ddP, nameP, &nd, 1);
  if (IS_ERR(tdP))
  {
    rc = PTR_ERR(tdP);
    code = 3;
    goto out_sdput;
  }

  rc = VFS_LINK(vfsP, sdP, ddP->d_inode, tdP);
  if (rc == 0)
    inode2pattr(sdP->d_inode, pattrP);
  dput(tdP);

out_sdput:
  dput(sdP);
out_ddput:
  PATH_PUT(&nd);

out:
  TRACE4(TRACE_PCACHE, 4, TRCID_CXICACHE_LINK,
         "cxiCacheLink: serialno %ld, new name %s, rc %d code %d",
         pattrP->pa_ino, nameP, rc, code);
  return -rc;
}

/* Create a symlink to an object identified by name and parent's fh and attrs */
int
cxiCacheSymlink(void *vfsP, pcacheAttr_t *parentPattrP,
                char *targetNameP, char *newNameP, pcacheAttr_t *cpattrP)
{
  struct dentry *ddP = NULL;
  struct dentry *dP = NULL;
  struct nameidata nd = {};
  int umask;
  int code = 0, rc = 0;

  /* Get dentry of parent */
  ddP = nfs_dget(vfsP, parentPattrP);
  if (IS_ERR(ddP))
  {
    rc = PTR_ERR(ddP);
    code = 1;
    goto out;
  }
  DBGASSERTRC(S_ISDIR(ddP->d_inode->i_mode), ddP->d_inode->i_ino, 0, 0);

  ND_PATH(nd).mnt = mntget(vfsP);
  ND_PATH(nd).dentry = ddP; /* We already have a hold on ddP */

  if (newNameP == NULL || *newNameP == '\0')
    /* We don't have the source name, so can't call cache_lookup */
    dP = nfs_dget(vfsP, cpattrP);
  else
    dP = cache_lookup(vfsP, ddP, newNameP, &nd, 1);

  if (IS_ERR(dP))
  {
    rc = PTR_ERR(dP);
    code = 2;
    goto out_put;
  }

  /* Use umask of 0 instead of process umask */
  umask = current->fs->umask;
  current->fs->umask = 0;
  rc = VFS_SYMLINK(vfsP, ddP->d_inode, dP, targetNameP);
  /* Reset umask */
  current->fs->umask = umask;

  if (rc == 0 && cpattrP != NULL)
    inode2pattr(dP->d_inode, cpattrP);
  dput(dP);

out_put:
  PATH_PUT(&nd);

out:
  TRACE5(TRACE_PCACHE, 4, TRCID_CXICACHE_SYMLINK,
         "cxiCacheSymlink: dir %lld target name %s new name %s, rc %d code %d",
         parentPattrP->pa_ino, targetNameP, newNameP, rc, code);
  return -rc;
}

/* Get attributes on an object identified by fh and attrs
   Use a flag to optionally force an update of attributes from the
   remove server */
int
cxiCacheGetattr(void *vfsP, pcacheAttr_t *pattrP, Boolean forceFlag)
{
  struct inode *iP = NULL;
  int rc = 0;
  struct kstat stat;
  struct dentry dentry;

  /* Get inode of object */
  iP = nfs_iget(vfsP, pattrP, true);
  if (IS_ERR(iP))
  {
    rc = PTR_ERR(iP);
    goto out;
  }

  dentry.d_inode = iP;

  /* Force attr invalid to refresh from server */
  if (forceFlag)
    NFS_CACHE_VALIDITY_FLAGS(iP) |= NFS_INO_INVALID_ATTR;
  rc = vfs_getattr(vfsP, &dentry, &stat);
  if (rc == 0)
    inode2pattr(iP, pattrP);

  iput(iP);

 out:
  TRACE2(TRACE_VNODE, 4, TRCID_CACHE_GETATTR,
         "cxiCacheGetattr: ino %lld rc %d",
         pattrP->pa_ino, rc);
  return -rc;
}

/* Set attributes on an object identified by fh and attrs.
 * Refresh pattrP with latest attributes. */
int
cxiCacheSetattr(void *vfsP, pcacheAttr_t *pattrP, int attrset, long arg1, long arg2)
{
  struct dentry *dP = NULL;
  struct iattr iattr;
  int rc = 0;

  /* Get inode of object */
  dP = nfs_dget(vfsP, pattrP);
  if (IS_ERR(dP))
  {
    rc = PTR_ERR(dP);
    goto out;
  }
  TRACE6(TRACE_PCACHE, 4, TRCID_CXICACHE_SETATTR_0,
         "cxiCacheSetattr: ino %lld attrset 0x%X mode %o uid %d gid %d size %ld",
         pattrP->pa_ino, attrset, pattrP->pa_mode,
         pattrP->pa_uid, pattrP->pa_gid, pattrP->pa_size);

  iattr.ia_valid = attrset;
  if (attrset == (ATTR_MODE | ATTR_SIZE | ATTR_UID | ATTR_GID))
  {
    /* called during recovery - pick attributes from attr */
    iattr.ia_mode = pattrP->pa_mode;
    iattr.ia_size = pattrP->pa_size;
    iattr.ia_uid = pattrP->pa_uid;
    iattr.ia_gid = pattrP->pa_gid;
  }
  else if (attrset & ATTR_MODE)
    iattr.ia_mode = (umode_t)arg1;
  else if (attrset & ATTR_SIZE)
    iattr.ia_size = (loff_t)arg1;
  else if (attrset & (ATTR_UID|ATTR_GID))
  {
    if (attrset & ATTR_UID)
      iattr.ia_uid = (uid_t)arg1;
    if (iattr.ia_valid & ATTR_GID)
      iattr.ia_gid = (gid_t)arg2;
  }

  /* FIX: we really need to call notify_change() here but it requires a
     valid dentry so dnotify_parent() doesn't oops the kernel. For now,
     it appears sufficient to just call setattr instead */
  rc = dP->d_inode->i_op->setattr(dP, &iattr);
  if (rc == 0)
  {
    /* Update attributes that _may_ be returned on a successful setattr (nfsv4) */
    inode2pattr(dP->d_inode, pattrP);
  }
  dput(dP);

 out:
  TRACE5(TRACE_PCACHE, 4, TRCID_CXICACHE_SETATTR,
         "cxiCacheSetattr: ino %lld attrset %d args 0x%lX 0x%lX rc %d",
         pattrP->pa_ino, attrset, arg1, arg2, rc);
  return -rc;
}

/* Close a file */
int
cxiCacheClose(void *fP, pcacheAttr_t *pattrP)
{
  struct file *fileP = (struct file *)fP;
  struct inode *inodeP = fileP->f_dentry->d_inode;
  int rc = 0;

  if (!igrab(inodeP))
    DBGASSERT(0);

  rc = filp_close(fileP, current->files);

  if (pattrP != NULL)
    inode2pattr(inodeP, pattrP);

  iput(inodeP);

  TRACE3(TRACE_PCACHE, 4, TRCID_CXICACHE_CLOSE,
         "cxiCacheClose: fP 0x%lX name %s, rc %d",
         fileP, (fileP->f_dentry) ? (char *)fileP->f_dentry->d_name.name : "null", rc);
  return -rc;
}

/* Perform READDIR operation on a dir */
int
cxiCacheReaddir(void *fP, void *fnP, void *argP)
{
  struct file *fileP = (struct file *)fP;
  int rc;

  rc = vfs_readdir(fileP, fnP, argP);
  TRACE5(TRACE_PCACHE, 4, TRCID_CXICACHE_READDIR,
         "cxiCacheReaddir: fP 0x%lX name %s flags 0%o mode %d rc %d",
         fileP, fileP->f_dentry->d_name.name, fileP->f_flags, fileP->f_mode, rc);
  return -rc;
}

/* Read a file (already open) starting at offsetP */
int cxiCacheRead(void *fP, char *bufP, size_t bufSize, offset_t *offsetP,
                 ssize_t *bytesRead, Boolean isUserBuf)
{
  struct file *fileP = (struct file *)fP;
  ssize_t rc = 0;
  int err = 0;

  DBGASSERT(fileP != NULL);
  *bytesRead = 0;

  if (isUserBuf)
    rc = vfs_read(fileP, (char __user *)bufP, bufSize, offsetP);
  else
    rc = kernel_read(fileP, *offsetP, bufP, bufSize);

  if (rc > 0)
  {
    *offsetP += rc;
    *bytesRead = rc;
    fileP->f_pos = *offsetP;
  }
  else
    err = -rc;

  TRACE7(TRACE_PCACHE, 4, TRCID_CXICACHE_READ,
         "cxiCacheRead: fileP 0x%lX flags 0%o mode %d offset %lld desired %lld read %lld rc %d",
         fileP, fileP->f_flags,
         fileP->f_mode, *offsetP, bufSize, *bytesRead, err);
  return err;
}

/* Write to a file (already opened) starting at offsetP
   Return updated attributes */
int cxiCacheWrite(void *fP, char *bufP, size_t bufSize, offset_t *offsetP,
                  ssize_t *bytesWritten, Boolean isUserBuf)
{
  struct file *fileP = (struct file *)fP;
  mm_segment_t old_fs;
  loff_t pos = *offsetP;
  ssize_t rc = 0;
  int err = 0;

  DBGASSERT(fP != NULL);
  if (isUserBuf)
    rc = vfs_write(fileP, (void __user *)bufP, bufSize, &pos);
  else
  {
    old_fs = get_fs();
    set_fs(get_ds());
    /* The cast to a user pointer is valid due to the set_fs() */
    rc = vfs_write(fileP, (void __user *)bufP, bufSize, &pos);
    set_fs(old_fs);
  }

  *bytesWritten = 0;
  if (rc > 0)
  {
    *offsetP += rc;
    *bytesWritten = rc;
    fileP->f_pos = pos;
  }
  else
    err = -rc;

  TRACE7(TRACE_PCACHE, 4, TRCID_CACHE_CXIWRITE,
         "cxiCacheWrite: fileP 0x%lX flags 0%o mode %d offset %lld desired %lld written %lld rc %d",
         fileP, fileP->f_flags,
         fileP->f_mode, *offsetP, bufSize, *bytesWritten, err);
  return err;
}

/* Rename a file, given names, fh's and attrs */
int
cxiCacheRename(void *vfsP, pcacheAttr_t *sDirPattrP, pcacheAttr_t *sPattrP, char *sNameP,
               pcacheAttr_t *tDirPattrP, pcacheAttr_t *tPattrP, char *tNameP)
{
  struct dentry *sdP = NULL, *sddP = NULL, *tdP = NULL, *tddP = NULL;
  struct nameidata nd = {};
  int rc = 0, code = 0;

  sddP = nfs_dget(vfsP, sDirPattrP);
  if (IS_ERR(sddP))
  {
    rc = PTR_ERR(sddP);
    code = 1;
    goto out;
  }
  DBGASSERTRC(S_ISDIR(sddP->d_inode->i_mode), sddP->d_inode->i_ino, 0, 0);

  ND_PATH(nd).mnt = mntget(vfsP);
  ND_PATH(nd).dentry = sddP; 

  sdP = cache_lookup(vfsP, sddP, sNameP, &nd, 0);
  if (IS_ERR(sdP))
  {
    rc = PTR_ERR(sdP);
    code = 3;
    goto out_sddput;
  }
  DBGASSERT(sddP == sdP->d_parent);

  if (tDirPattrP != NULL)
  {
    tddP = nfs_dget(vfsP, tDirPattrP);
    if (IS_ERR(tddP))
    {
      rc = PTR_ERR(tddP);
      code = 2;
      goto out_sdput;
    }
    DBGASSERTRC(S_ISDIR(tddP->d_inode->i_mode), tddP->d_inode->i_ino, 0, 0);
    ND_PATH(nd).dentry = tddP;
  }
  else
    tddP = sddP;

  tdP = cache_lookup(vfsP, tddP, tNameP, &nd, 1);
  if (IS_ERR(tdP))
  {
    code = 6;
    rc = PTR_ERR(tdP);
    goto out_tddput;
  }

  rc = VFS_RENAME(vfsP, sddP->d_inode, sdP, tddP->d_inode, tdP);

  /* Hack: there is a problem with renames of directories where
     the target dir exists and the dentry name is not inline.
     A subsequent lookup finds the dentry by name although it should
     have been unhashed. Until this problem is found, clear the name
     so it's not found by a subsequent lookup */
  if (rc == 0 && tPattrP != NULL && S_ISDIR(sdP->d_inode->i_mode) &&
      sdP->d_name.len >= DNAME_INLINE_LEN_MIN)
    cxiMemcpy((void *)tdP->d_name.name, (void *)sdP->d_name.name, tdP->d_name.len);

  dput(tdP);

out_tddput:
  if (tddP && tddP != sddP)
    dput(sddP);
out_sdput:
  dput(sdP);
out_sddput:
  PATH_PUT(&nd);

out:
  TRACE6(TRACE_PCACHE, 4, TRCID_CXICACHE_RENAME,
         "cxiCacheRename: source %s target %s srcdir %lld tgtdir %lld, rc %d code %d",
         sNameP, tNameP, sDirPattrP->pa_ino, (tDirPattrP != NULL) ? tDirPattrP->pa_ino : -1LL,
         rc, code);
  return -rc;
}

/* Perform READLINK operation given the attributes and target name */
int
cxiCacheReadlink(void *vfsP, pcacheAttr_t *pattrP, char *targetP)
{
  struct dentry *dentryP = NULL;
  struct nameidata nd = {};
  mm_segment_t oldfs;
  char *linkP = (char *) __get_free_page(GFP_USER);
  int len = 0, rc = 0;

  if (linkP == NULL)
  {
    rc = -ENOMEM;
    goto out;
  }

  dentryP = nfs_dget(vfsP, pattrP);
  if (IS_ERR(dentryP))
  {
    rc = PTR_ERR(dentryP);
    goto out_free;
  }

  oldfs = get_fs(); set_fs(KERNEL_DS);
  len = dentryP->d_inode->i_op->readlink(dentryP, linkP, PAGE_SIZE);
  set_fs(oldfs);

  if (len < 0)
    rc = len;
  else if (len > CXI_PATH_MAX)
    /* FIX: we don't support symlinks > CXI_PATH_MAX */
    rc = -ENAMETOOLONG;
  else
  {
    strncpy(targetP, linkP, len);
    targetP[len] = '\0';
    rc = 0;
  }
  dput(dentryP);

 out_free:
  free_page((unsigned long)linkP);
 out:
  TRACE3(TRACE_PCACHE, 4, TRCID_CXICACHE_READLINK,
         "cxiCacheReadLink: targetP %s rc %d len %d", targetP, rc, len);
  return -rc;
}

/*
 * Other helper routines 
 */

/* Get parent inode and file name used at open time from struct file */
int cxiCacheOpenState(void *dP, cxiNode_t **cnP, char *nameP)
{
  struct inode *inodeP = NULL;
  struct dentry *dentryP = dP;
  int rc = -ENOENT;

  if (dentryP == NULL)
    goto out;

  if (IS_ROOT(dentryP))
  {
    *nameP = '.';
    *((char *)nameP+1) = '\0';
  }
  else
  {
    char *dnameP = (char *)dentryP->d_name.name;
    if (dnameP == NULL || *dnameP == '\0' || dentryP->d_parent == NULL)
      goto out;
    cxiStrcpy(nameP, dnameP);
  }

  inodeP = (void *)dentryP->d_parent->d_inode;
  if (inodeP == NULL)
    goto out;

  *cnP = VP_TO_CNP(inodeP);
  rc = 0;

out:
  TRACE3(TRACE_PCACHE, 4, TRCID_CXICACHE_OPEN_EXIT,
         "cxiCacheOpenState: dentryP 0x%lX name %s rc %d", dentryP, nameP, rc);
  return -rc;
}

/* Obtain the number of blocks and file size for a given inode.
   Also return NFS read size */
int
cxiDataSize(void *vfsP, pcacheAttr_t *pattrP,
            UInt64 *dataSizeP, UInt64 *fileSizeP, UInt32 *readSizeP)
{
  struct vfsmount *mntP = (struct vfsmount *)vfsP;
  struct inode *iP = NULL;
  int rc = 0;

  *readSizeP = NFS_SERVER(mntP->mnt_root->d_inode)->rsize;
  if (pattrP == NULL)
    goto out;

  /* Get inode of object */
  iP = nfs_iget(vfsP, pattrP, false);
  if (IS_ERR(iP))
  {
    rc = PTR_ERR(iP);
    goto out;
  }

  /* Number of block is for blocks that are 512 bytes in size */
  *dataSizeP = iP->i_blocks * 512; /* GET_INODE_BLOCKSIZE(iP);*/
  *fileSizeP = iP->i_size;

  TRACE6(TRACE_VNODE, 3, TRCID_SET_DATASIZE,
         "cxiDataSize: ino %lld blksize 0x%X blocks %d dsize %llu fsize %llu rsize %u",
         iP->i_ino, GET_INODE_BLOCKSIZE(iP), iP->i_blocks,
         *dataSizeP, *fileSizeP, *readSizeP);
  iput(iP);

 out:
  return -rc;
}

void
cxiDeleteDCacheEntry(void *dP)
{
  struct dentry *dentryP = (struct dentry *)dP;
  DBGASSERT(dentryP != NULL);

  TRACE3(TRACE_VNODE, 4, TRCID_DEL_DCACHE_ENTRY,
         "cxiDeleteDCacheEntry: dentry 0x%lX d_inode 0x%lX (name '%s')\n",
         dentryP, dentryP->d_inode, dentryP->d_name.name);

  d_delete(dentryP);
}

/* Open a file over NFS given its parent attributes and name
   and return a file descriptor (that can be used in user space to
   do regular fs operations). The caller must call close(fd).
   mntPathP points to the root of the NFS mount */
int
cxiCacheNfsOpen(const char *mntPathP, pcacheAttr_t *pattrP,
                const char *nameP, int flags, int *fdP)
{
  void *vfsP = NULL;
  struct file *fileP = NULL;
  loff_t fileSize = 0;
  int fd = get_unused_fd();
  int rc = 0, code = 0;

  if (fd < 0)
  {
    rc = fd;
    code = 1;
    goto out;
  }

  rc = cxiCacheInit(mntPathP, NULL, &vfsP);
  if (rc != 0)
  {
    code = 2;
    goto out;
  }

  rc = cxiCacheOpen(vfsP, pattrP, nameP, flags,
                    0, 0, 0, 
                    NULL, (void *)&fileP, &fileSize);
  if (rc != 0)
  {
    code = 3;
    goto out;
  }

  /* fsnotify_open(fileP->f_dentry); */
  fd_install(fd, fileP);
  *fdP = fd;

out:
  if (vfsP != NULL)
    (void)cxiCacheExit(vfsP, NULL); /* ignore error */

  TRACE7(TRACE_PCACHE, 4, TRCID_CACHE_NFS_OPEN,
         "cxiCacheNfsOpen: path %s parent ino %lld name %s flags %d "
         "fd %d code %d rc %d",
         mntPathP, pattrP->pa_ino, nameP, flags, fd, code, rc);
  return rc;
}

/* Return the filename given inode numbers and a fold value */
int
cxiCacheGetFilename(Boolean pcache, void *vfsP, unsigned long pino,
                    unsigned long cino, char *nameP, int foldVal)
{
  int rc = ENOENT, code = 0;
  struct super_block *sb = (struct super_block *)vfsP;
  struct dentry *pDentry = NULL, *cDentry = NULL;
  unsigned int genNum = 0xffffffff;

  pDentry = get_dentry(sb, pino, genNum);
  if (pDentry == NULL || IS_ERR(pDentry))
  {
    TRACE2(TRACE_PCACHE, 4, TRCID_CACHE_GET_FNAME_1,
           "cxiCacheGetFilename pDentry 0x%X err %d",
            pDentry, PTR_ERR(pDentry));
    code = 1;
    goto out;
  }
  cDentry = get_dentry(sb, cino, genNum);
  if (cDentry == NULL || IS_ERR(cDentry))
  {
    TRACE2(TRACE_PCACHE, 4, TRCID_CACHE_GET_FNAME_2,
           "cxiCacheGetFilename cDentry 0x%X err %d",
           cDentry, PTR_ERR(cDentry));
    dput(pDentry);
    code = 2;
    goto out;
  }

  rc = gpfs_get_name_by_fold(pcache, pDentry, nameP, cDentry, foldVal);
  dput(pDentry);
  dput(cDentry);

out:
  TRACE6(TRACE_PCACHE, 4, TRCID_CACHE_GET_FNAME,
         "cxiCacheGetFilename: sb 0x%X pino %lld cino %lld name %s code %d rc %d",
         sb, pino, cino, nameP, code, rc);
  return -rc;
}

/* Return the next filename given parent inode number for hard links */
int
cxiCacheGetNextFilename(void *vfsP, unsigned long pino,
                        getdents_callback_t *getDirEntP)
{
  int rc = ENOENT, code = 0;
  struct super_block *sb = (struct super_block *)vfsP;
  struct dentry *pDentry = NULL;
  unsigned int genNum = 0xffffffff;
  struct getdents_callback_s buffer;

  TRACE5(TRACE_PCACHE, 4, TRCID_CACHE_GET_NEXTFNAME_0,
         "cxiCacheGetNextFilename pino %lld cio %lld found %d seq %d pos %lld",
          pino, getDirEntP->ino, getDirEntP->found, getDirEntP->sequence,
          getDirEntP->pos);

  pDentry = get_dentry(sb, pino, genNum);
  if (pDentry == NULL || IS_ERR(pDentry))
  {
    TRACE2(TRACE_PCACHE, 4, TRCID_CACHE_GET_NEXTFNAME_1,
           "cxiCacheGetNextFilename pDentry 0x%X err %d",
            pDentry, PTR_ERR(pDentry));
    code = 1;
    goto out;
  }

  rc = slow_get_next_name(pDentry, getDirEntP);
  dput(pDentry);

out:
  TRACE9(TRACE_PCACHE, 4, TRCID_CACHE_GET_NEXTFNAME,
         "cxiCacheGetNextFilename: sb 0x%X pino %lld cino %lld name %s "
         "found %d seq %d pos %lld code %d rc %d",
         sb, pino, getDirEntP->ino, getDirEntP->name,
         getDirEntP->found, getDirEntP->sequence, getDirEntP->pos, code, rc);
  return -rc;
}

int
cxiCacheFadvise(void* fP, int advice, loff_t len)
{
  int rc = 0;
  struct file *fileP = (struct file *)fP;
  struct address_space *mappingP = fileP->f_mapping;
  struct backing_dev_info *bdiP = NULL;
  loff_t endbyte = 0, offset = 0;
  pgoff_t start_index = 0, end_index = 0;
  unsigned long nrpages = 0;

  if (!fileP)
  {
    rc = EBADF;
    goto out;
  }

  if (S_ISFIFO(fileP->f_dentry->d_inode->i_mode))
  {
    rc = ESPIPE;
    goto out;
  }

  if (!mappingP)
  {
    rc = EINVAL;
    goto out;
  }

  /* len == 0 means "as much as possible" */
  endbyte = offset + len;
  if (!len || endbyte < len)
	  endbyte = -1;
  else
	  endbyte--;

  bdiP = mappingP->backing_dev_info;

  switch(advice)
  {
  case POSIX_FADV_SEQUENTIAL:
    fileP->f_ra.ra_pages = bdiP->ra_pages * 2;
#if LINUX_KERNEL_VERSION >= 2063200
    spin_lock(&fileP->f_lock);
    fileP->f_mode &= ~FMODE_RANDOM;
    spin_unlock(&fileP->f_lock);
#endif
    break;
#if LINUX_KERNEL_VERSION >= 2061800
  case POSIX_FADV_DONTNEED:
    if (!bdi_write_congested(mappingP->backing_dev_info))
      filemap_flush(mappingP);
 
    /* First and last FULL page! */
    start_index = (offset + (PAGE_CACHE_SIZE-1)) >> PAGE_CACHE_SHIFT;
    end_index = (endbyte >> PAGE_CACHE_SHIFT);
 
    if (end_index >= start_index)
      invalidate_mapping_pages(mappingP, start_index, end_index);
    break;
#endif
#if LINUX_KERNEL_VERSION >= 2063200
  case POSIX_FADV_WILLNEED:
/*
    start_index = offset >> PAGE_CACHE_SHIFT;
    end_index = endbyte >> PAGE_CACHE_SHIFT;
    nrpages = end_index - start_index + 1;
    if (!nrpages)
      nrpages = ~0UL;
    rc = force_page_cache_readahead(mappingP, fileP, start_index, nrpages);
*/
    page_cache_sync_readahead(mappingP, &fileP->f_ra, fileP, offset, len);

    if (rc > 0)
      rc = 0;
    break;
#endif
  default:
    rc = EINVAL;
  }

out:
  TRACE4(TRACE_PCACHE, 6, TRCID_CACHE_FADVISE,
         "cxiCacheFadvise: fp 0x%X advice %d len %lld rc %d",
         fP, advice, len, rc);

  return -rc;
}

