/***************************************************************************
 *
 * Copyright (C) 2001 International Business Machines
 * All rights reserved.
 *
 * This file is part of the GPFS mmfslinux kernel module.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer. 
 *  2. Redistributions in binary form must reproduce the above copyright 
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution. 
 *  3. The name of International Business Machines may not be used to endorse 
 *     or promote products derived from this software without specific prior 
 *     written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR 
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Alternatively, provided that this notice is retained in full, this
 * software may be distributed under the terms of the GNU General
 * Public License ("GPL") as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any later version,
 * in which case the provisions of the GPL apply INSTEAD OF those given above.
 *
 ****************************************************************************/
 static char sccsid[] = "@(#)37	1.106.1.8  src/avs/fs/mmfs/ts/kernext/gpl-linux/lxtrace.c, mmfs, avs_rhin, rhins014a 5/21/12 19:06:58";
/**************************************************************************
 *
 * lxtrace : the command.  Turns tracing on/off and formats trace files.
 *
 *  "lxtrace {on hw[,hw[,hw...]] [trcFile [filesize [bufsize]]]" }
 *           {off                                               }
 *           {format [-T] [-v] [-t formatFile] [-o outFile] [trcFile] }
 *
 *           on   - begins tracing
 *           off  - stops tracing
 *           hw   - hookword (identifies traces to be captured)
 *                  This is a comma-separated list (up to LXTRACE_MAX_HW)
 *           trcFile  - path to the file where output is written
 *           filesize - maximum size (bytes) of the wrap-around file
 *                      A filesize of 0 will cause the file NOT to wrap
 *           bufsize  - the size of buffer to used by the device
 *                      to record trace records
 *
 *   For debug/testing purposes, the following additional options
 *   are defined:
 *   "lxtrace dump              // returns state information in a buffer
 *           fsync              // flush buffered records to the read buffer
 *
 **************************************************************************/
#ifdef GPFS_LINUX
/* Added for LFS */
#define _LARGEFILE_SOURCE
#define _FILE_OFFSET_BITS 64
#include <Shark-gpl.h>
#else
#include <cxiTypes.h>          // needed by windows per Demyn
#endif

#include <sys/types.h>
#include <sys/time.h>
#include <sys/ioctl.h>

#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <malloc.h>
#include <signal.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/stat.h>
#include <dlfcn.h>
#include <limits.h>

#include <Trace.h>
#define ROUNDUP(_addr, _align) \
   (((long)_addr + (long)_align - 1L) & ~((long)_align - 1L))
#include <lxtrace.h>

#if defined(GPFS_LINUX) && defined(LXTRACE_USE_RELAY)
extern int RelayDump(void);
extern int RelayStart(long long outSize, int bufSize, int bufNum, int interval, int level);
#endif
static int (*uncompress)(char *dst, unsigned long *dlen, char *src, unsigned long slen);
extern int msgGet(int flags);
extern unsigned int cpuNum;
extern size_t compLevel;
extern int useDIO;
extern long long percpuOutSize;
extern size_t subBufSize;
extern int writeMode;
extern int noWrap;
extern char percpuOutBasename[PATH_MAX];
/* Examples of formatted trace output:

   Default options:

    Timestamp   Pid  COMPONENT_TAG: application trace record
   ----------- ----- ---------------------------------------
      0.000000 21499 TRACE_VNODE: gpfs_i_revalidate enter:

   -v option:

    Timestamp   Pid  P Hookword COMPONENT_TAG: application trace record
   ----------- ----- - -------- ---------------------------------------
      0.000000 21499 1 307004B3 TRACE_VNODE: gpfs_i_revalidate enter:

   -T option:

      Timestamp      Pid  COMPONENT_TAG: application trace record
   ---------------- ----- ---------------------------------------
   970706704.711164 21499 TRACE_VNODE: gpfs_i_revalidate enter:

   -T and -v options:

      Timestamp      Pid  P Hookword COMPONENT_TAG: application trace record
   ---------------- ----- - -------- ---------------------------------------
   970706704.711164 21499 1 307004B3 TRACE_VNODE: gpfs_i_revalidate enter:

 */

/* Define fragments of header lines and the corresponding trace printf format
   strings for each field on a trace output line */
#define H1_TS       " Timestamp  "
#define H2_TS       "----------- "
#define FMT_TS      "%4d.%06d "
#define H1_TS_T     "    Timestamp     "
#define H2_TS_T     "----------------- "
#define FMT_TS_T    "%10d.%06d "

#ifdef GPFS_SOLARIS
#define H1_PID      "  Pid  :  Tid  "
#define H2_PID      "-------------- "
#define FMT_PID     "%6d : %5lu "
#else
#define H1_PID      "  Pid  "
#define H2_PID      "------ "
#define FMT_PID     "%6d "
#endif

#define H1_SEQ      " Sequence  "
#define H2_SEQ      "---------- "
#define FMT_SEQ     "%10llu "

#define H1_TS_D     "           Date            "
#define H2_TS_D     "-------------------------- "
#define FMT_TS_D    "%19s.%06d "

#define H1_P_HW     ""
#define H2_P_HW     ""
#define FMT_P_HW    ""
#define H1_P_HW_V   "P B Hookword "
#define H2_P_HW_V   "- - -------- "
#define FMT_P_HW_V  "%1d %1d %08X "

#define H1_TAG      "COMPONENT_TAG: application trace record\n"
#define H2_TAG      "---------------------------------------\n"
#define FMT_TAG     "%s: "

/* Bytes in an "_id2name" formatting substitution */

#define FMT_TRC_BEGIN "Trace begins: %s\n"
#define FMT_TRC_END "Trace ends  : %s%s\n"
#define FMT_COMP_RATIO "Trace compression ratio: %2d\n"
#define FMT_LOST_COUNT "Trace records lost in total: %10llu\n"

#define TRC_BEGIN_STR "formatting...           "
#define TRC_END_STR   "formatting...           ","               "
#define FILE_WRAP_STR " (file wrapped)"
#define FILE_NOWRAP_STR "               "

#if defined( __64BIT__) || defined(_LP64)
typedef long long argType;
#define ARGLEN sizeof(long long)
#define TRC_ARG(N) (((long long *)&application_data)[N])
#else
typedef int argType;
#define ARGLEN sizeof(int)
#define TRC_ARG(N) (((int *)&application_data)[N])
#endif // __64BIT_ || _LP64_

#define TRC_ARGLIST argP[0] , argP[1] , argP[2] , argP[3] , argP[4] , argP[5], \
                    argP[6] , argP[7] , argP[8] , argP[9] , argP[10], argP[11],\
                    argP[12], argP[13], argP[14], argP[15], argP[16], argP[17],\
                    argP[18], argP[19]


/* Offsets of the various fields in the formatting file */
#define GFR_TRCID 0
#define GFR_HOOKID 15
#define GFR_COMPID 24
#define GFR_FORMAT 44

/* Max length of COMPID in the formatting file */
# define MAX_COMPID_LEN (GFR_FORMAT - GFR_COMPID - 1)

/* Thread IDs greater than this are really message handlers.
   FirstMsgHandlerThread is also defined in ThreadNames.h. */
#define FirstMsgHandlerThread 0x200

/* Size of the hash tables for in-memory format file and its hash functions.
   HOOK_HASH is for looking up format strings by hook id.
   I2N_HASH is for looking up id2name conversions by conversion name and id
   value; the I2NHASH hash function uses the first 4 chars of the conversion
   name, relying on the fact that these names are at least 4 chars long,
   because they all end in "_id2name". */
#define HOOK_HASH_TABLE_SIZE 0x100000
/* The full hookid values for hookhash are of the form YYY0XXXX, where Y is the
   sharkhookid and X is the trace id. The sharkhookids are one of 30[6-9], so only
   the last digit of the hookid is important. */
int hookhash(int H)
{
  int hookid = ((H & 0x00F00000) >> 4); //we don't need the 0 in the middle
  int trcid = (H & 0x0000FFFF);
  return (hookid | trcid);
}
#define I2N_HASH_TABLE_SIZE 0x4000
#ifndef ALIGNED_DATA_ACCESS
#define I2NHASH(F,I) ((*(int*)(F) + (I)) & 0x3FFF)
#else
static inline int I2NHASH(char *F, int I)
{
  int tmp;

  memcpy(&tmp, F, sizeof(int));
  return (tmp + I) & 0x3FFF;
}
#endif

/* combined device and application trace data headers */
typedef struct trc_hdrs_t
{
  trc_header_t hdr;  /* device header */
  trc_datahdr_t rec; /* application header */
} trc_hdrs_t;

/* Each format record has one of these headers describing it. */
typedef struct fmtDesc_t
{
  int       hookid;             // hookid or id2name value
  char      *cidP;              // component identifier or id2name tag
  char      *fmtP;              // format string or id2name string
  unsigned int llmask;          // bitvector indicating %ll arguments
  i2nSpec_t *i2nSpecP;          // array of id2name conversion specifiers
  struct fmtDesc_t *fmtNextP;   // next in the hash chain
  int numHits;                  //number of times we see this hookid
  Int64 totalSpace;             //total space in the trcfile for this hookid
  int largestEntry;             //the largest entry (by space) for this hookid in the trcfile
} fmtDesc_t;


/* Hash tables for finding format records. */
fmtDesc_t * hookHashP[HOOK_HASH_TABLE_SIZE];
fmtDesc_t * i2nHashP[I2N_HASH_TABLE_SIZE];

/* Hash lookup statistics */
static int nHookLookup = 0;
static int nHookCheck = 0;
static int nI2NLookup = 0;
static int nI2NCheck = 0;
static int nI2NFail = 0;

/* static state information */
int TraceHandle = -1; // file descriptor for the trace device
static long long outSize;       // out_fd filesize
static int bufSize;       // size of the trace buffers (device and daemon)
static int bufNum;
static int interval;
static int useRelay;
static char *bufP;        // the daemon trace read buffer
static int hdrSize[LXTRACE_NR_CPUS];  // trace header size, for backward compatibility
unsigned long long totalPostCompSize = 0;
unsigned long long totalTrcfileProcessedSize = 0; //I need to track this seperately since overwrite mode doesn't do compression
unsigned long long totalCompSize = 0;

#if !defined(GPFS_SOLARIS)
/* this code is used on Linux and Solaris, but only allow DIO on Linux */
#define LXTRACE_USE_DIO
#endif
static int nWraps;        // number of times the output file has wrapped
static off_t currentPos;  // position of next write to output file
static off_t fileEnd=0;  // position of next write to output file
static int curCPU;
extern int msgQueue;
extern int relayBuf;
static int genStats = 0;
/* The previous time is used to validate a wrap point when we
   think we've encountered one. */
static struct timeval prevTime[LXTRACE_NR_CPUS];
/* Keep track of start/end times in the trace file */
static time_t startTime, stopTime;

#ifndef GPFS_WINDOWS
static char *formatFileP;     // malloc'ed buffer holding formatting data
static char *fmtDescSpaceP;   // fmtData_t space for each format record
#else
extern char *formatFileP;     // malloc'ed buffer holding formatting data
extern char *fmtDescSpaceP;   // fmtData_t space for each format record
#endif

static int allowed_usec_wobble = 2; // allowed timestamp wobble before declaring a wrap.

/* Convert ascii representation of a hexadecimal to int */
int axtoi(char s[])
{
  int i, n;

  n = 0;
  for (i = 0; s[i] != '\0'; i++)
  {
    if (s[i] >= '0' && s[i] <= '9')
      n = 16 * n + (s[i] - '0');
    else if (s[i] >= 'A' && s[i] <= 'F')
      n = 16 * n + (s[i] - 'A' + 10);
    else if (s[i] >= 'a' && s[i] <= 'f')
      n = 16 * n + (s[i] - 'a' + 10);
    else
      break;
  }
  return n;
}


/* Parse a quoted format string: removes enclosing quotes, removes trailing
   white-space, and converts escape sequences \n, \r, \t, \", etc.  Also
   examines '%' format arguments and computes a bitvector that indicates which
   of the arguments are long longs (%ll..).  The converted string is placed in
   outP; the i'th bit in *llmaskP is set if the i'th argument is a long long.
   It's ok to call this function to convert a string in-place (outP == inP),
   since the conversion only makes the string shorter (removes surrounding
   quotes and reduces escape sequences to a single char).
   Returns a pointer to the next character in inP following the ending quotes,
   or NULL if there are no more characters following the quoted string. */
char *ParseFormatString(char *inP, char *outP, unsigned int *llmaskP)
{
  char *initialOutP = outP;
  int b = 1;

  /* skip initial quote */
  if (*inP == '"')
    inP++;

  /* copy string until ending quote */
  *llmaskP = 0;
  while (*inP && *inP != '"')
  {
    if (*inP == '\\' && *(inP + 1))
    {
      inP++;
      switch (*inP)
      {
        case 'n': *outP++ = '\n'; inP++; break;
        case 'r': *outP++ = '\r'; inP++; break;
        case 't': *outP++ = '\t'; inP++; break;
        default:  *outP++ = *inP++;
      }
    }
    else if (*inP == '%')
    {
      *outP++ = *inP++;
      if (*inP == '%')
        *outP++ = *inP++;
      else
      {
        if (*inP == '+' || *inP == '-')
          *outP++ = *inP++;
        while (isdigit(*inP))
          *outP++ = *inP++;
        if (*inP == 'l' && *(inP+1) == 'l')
          *llmaskP |= b;
        b <<= 1;
      }
    }
    else
      *outP++ = *inP++;
  }

  /* remove trailing white space from the string */
  while (outP > initialOutP && isspace(outP[-1]))
    outP--;
  *outP = '\0';

  /* Return pointer to next character after the string or NULL if no ending
     quote was found */
  return *inP? inP + 1: NULL;
}


/* Parse a string containing a list of id2name conversion specifiers of the
   form "$n=xxx_id2name $m=yyy_id2name ...".  Returns a "null-terminted" array
   of i2nSpec_t structures, i.e., the array contains one more entry than the
   number of conversion specifiers found, and argNum/i2nNameP in the last
   array entry are set to 0/NULL. */
i2nSpec_t *ParseI2N(char *i2nP, unsigned int llmask)
{
  i2nSpec_t i2n[LXTRACE_MAX_FORMAT_SUBS];
  i2nSpec_t *i2nSpecP;
  char *p;
  int n, argNum, a, nll;
  unsigned int m;

  for (n = 0, p = strtok(i2nP, " \t\n");
       p && n < LXTRACE_MAX_FORMAT_SUBS;
       n++, p = strtok(NULL, " \t\n"))
  {
    if (*p != '$')
      break;
    p++;
    argNum = atoi(p);
    p = strchr(p, '=');
    if (argNum <= 0 || p == NULL)
      break;

#if !defined(__64BIT__) && !defined(_LP64)
    /* The "n" in "$n=" signifies that the "n"th parameter needs substitution.
       However, we need to account for any double-word ("%ll") values (will be
       taking two positions) in the arglist and adjust accordingly.  The
       llmask parameter indicates which arguments are of the %ll form.  Count
       the bits in llmask before the point of the substitution and add the
       count to the argNum value. */
    if (llmask)
    {
      nll = 0;
      for (a = 1, m = llmask;
           a < argNum && m;
           a++, m >>= 1)
      {
        if (m & 1)
          nll++;
      }
      argNum += nll;
    }
#endif // __64BIT__

    i2n[n].argNum = argNum;
    i2n[n].i2nNameP = p + 1;
  }
  if (n == 0)
    return NULL;  // ?? print msg about bad id2name specifiers

  i2nSpecP = (i2nSpec_t *)malloc((n + 1)*sizeof(i2nSpec_t));
  if (i2nSpecP != NULL)
  {
    memcpy(i2nSpecP, i2n, n*sizeof(i2nSpec_t));
    i2nSpecP[n].argNum = 0;
    i2nSpecP[n].i2nNameP = NULL;
  }
  return i2nSpecP;
}


/* The formatting records are read into memory and parsed.  A fmtDesc_t header
   is constructed for each and is then inserted into a hash table. */
int BuildFormatTable(FILE *fP)
{
  char buf[2048];
  long fileSize;   /* size (bytes) of the formatting file */
  int  nRecords;   /* Number of records in the formatting file */
  int  hookid;     /* hookid of current record */
  int  hashVal;    /* hash value */
  struct fmtDesc_t *nextDescP;
  char *curFmtRecordP, *nextFmtRecordP;
  char *trcidP;
  char *hookP;
  char *cidP;
  char *fmtP;
#ifdef DBGASSERTS
  char *formatFileEndP;
  char *fmtDescSpaceEndP;
#endif // DBGASSERTS

  /* Clear all entries in the hash table */
  memset(hookHashP, 0, sizeof(hookHashP));
  memset(i2nHashP,  0, sizeof(i2nHashP));

  rewind(fP);

  /* Count the number of formatting records */
  nRecords = 0;
  while (fgets(buf, sizeof(buf), fP) != NULL)
    nRecords += 1;

  /* Now that we are at the end of the file, see how big it is */
  fileSize = ftell(fP);
  if (fileSize < 0)
    return errno;

  /* Allocate the buffer for the file contents.
     Add 1 for possible trailing zero. */
  if (NULL == (formatFileP = (char *)malloc(fileSize+1)))
    return ENOMEM;

  /* Allocate space for the fmtData_t for all of the records. */
  if (NULL == (fmtDescSpaceP = (char *)malloc(nRecords * sizeof(fmtDesc_t))))
    return ENOMEM;
  nextDescP = (fmtDesc_t *)fmtDescSpaceP;

#ifdef DBGASSERTS
  formatFileEndP = formatFileP + fileSize;
  fmtDescSpaceEndP = fmtDescSpaceP + (nRecords * sizeof(fmtDesc_t));
#endif // DBGASSERTS

  rewind(fP); /* back to the beginning of the file */

  /* Read each record in one-at-a-time, parse the record, and insert it into
     the right hash table. */
  curFmtRecordP = formatFileP;
  while (fgets(curFmtRecordP, sizeof(buf), fP) != NULL)
  {
    nextFmtRecordP = curFmtRecordP + strlen(curFmtRecordP);

    /* The TRCID identifier, hookid, component id, and format string are at
       fixed offsets */
    trcidP = curFmtRecordP + GFR_TRCID;
    hookP  = curFmtRecordP + GFR_HOOKID;
    cidP   = curFmtRecordP + GFR_COMPID;
    fmtP   = curFmtRecordP + GFR_FORMAT;

    /* parse the format record. */
    hookid = axtoi(hookP) & 0xFFFFFFFF;
    nextDescP->hookid = hookid;
    nextDescP->cidP = strtok(cidP, " ");
    nextDescP->fmtP = fmtP;

    /* find the end of the format string */
    fmtP = ParseFormatString(fmtP, fmtP, &nextDescP->llmask);

    /* check for id2name conversion specifiers following the format string */
    nextDescP->i2nSpecP = NULL;
    if (fmtP)
    {
      while (isspace(*fmtP))
        fmtP++;
      if (*fmtP)
        nextDescP->i2nSpecP = ParseI2N(fmtP, nextDescP->llmask);
    }

    /* Place the format record descriptor into the right hash table. */
    if (strncmp(trcidP, "TRCID_", 6) == 0)
    {
      /* this record contains a format string */
      strcat(nextDescP->fmtP, "\n");
      hashVal = hookhash(hookid);
      nextDescP->fmtNextP = hookHashP[hashVal];
      hookHashP[hashVal] = nextDescP;
      if (genStats)
      {
        hookHashP[hashVal]->numHits = 0;
        hookHashP[hashVal]->totalSpace = 0;
        hookHashP[hashVal]->largestEntry = 0;
      }
    }
    else
    {
      /* this record contains an id2name conversion value */
      hashVal = I2NHASH(nextDescP->cidP, hookid);
      nextDescP->fmtNextP = i2nHashP[hashVal];
      i2nHashP[hashVal] = nextDescP;
    }

    /* Advance to the next fmtDesc and format record */
    nextDescP++;
    curFmtRecordP = nextFmtRecordP;
  }

#ifdef DBGASSERTS
  /* verify that we didn't overrun our buffers */
  if ((char *)nextDescP > fmtDescSpaceEndP)
  {
    DP("%s: overwrote format descriptor memory!\n",TRCFMT_NAME);
    return -1;
  }
  if (nextFmtRecordP > formatFileEndP)
  {
    DP("%s: overwrote format file memory!\n",TRCFMT_NAME);
    return -1;
  }
#endif // DBGASSERTS

  return 0;
}


/* Find a format record by hookid. */
int GetFormatRecord(FILE *fP, int hookid, char **cidPP, char **fmtPP,
                    unsigned int *llmaskP, i2nSpec_t **i2nPP,
                    Int64 spaceUsed)
{
  fmtDesc_t *nextP;

  DFP("GetFormatRecord(0x%X, 0x%X, 0x%X, 0x%X, 0x%X)\n",
      fP, hookid, bufSize, bufP, cidPP);

  nHookLookup++;

  /* search hash chain for the specified hookid */
  for (nextP = hookHashP[hookhash(hookid)];
       nextP != NULL;
       nextP = nextP->fmtNextP)
  {
    nHookCheck++;

    /* Check whether hookid matches. */
    if (nextP->hookid == hookid)
    {
      /* Hookid and the record type match.  Copy to work area. */
      *cidPP = nextP->cidP;
      *fmtPP = nextP->fmtP;
      *llmaskP = nextP->llmask;
      *i2nPP = nextP->i2nSpecP;
      if (genStats)
      {
        nextP->numHits++;
        nextP->totalSpace += spaceUsed;
        if (nextP->largestEntry < spaceUsed)
          nextP->largestEntry = spaceUsed;
        totalTrcfileProcessedSize += spaceUsed;
      }
      DFP("GetFormatRecord for id %X is: %s\n", hookid, *fmtPP);
      return 0;
    }
  }

  /* not found */
  return -1;
}

/* For genStats (--stat flag), sort the hookids for display */
int hookhashComparator(const void *p1, const void *p2)
{
  fmtDesc_t *a = *((fmtDesc_t  **) p1);
  fmtDesc_t *b = *((fmtDesc_t **) p2);

  if (a == NULL && b == NULL)
    return 0;
  else if (a == NULL)
    return -1;
  else if (b == NULL)
    return 1;

  //sort first by hits, then by total space
  if (a->numHits > b->numHits)
    return 1;
  else if (a->numHits < b->numHits)
    return -1;

  if (a->totalSpace > b->totalSpace)
    return 1;
  else if (a->totalSpace < b->totalSpace)
    return -1;

  return 0;
}

/* Find an id2name string by conversion name ("tag") and id value */
int GetID2NameRecord(FILE *fP, char *tagP,
                     int id, int bufSize, char *bufP)
{
  fmtDesc_t *nextP;

  DFP("GetID2NameRecord(0x%X, 0x%X, 0x%X, 0x%X, 0x%X)\n",
      fP, tagP, id, bufSize, bufP);

  nI2NLookup++;

  /* search hash chain for the specified hookid and tag */
  for (nextP = i2nHashP[I2NHASH(tagP, id)];
       nextP != NULL;
       nextP = nextP->fmtNextP)
  {
    nI2NCheck++;

    /* Check whether hookid and tag match. */
    if (nextP->hookid == id &&
        strncmp(nextP->cidP, tagP, MAX_COMPID_LEN) == 0)
    {
      DFP("GetID2NameRecord for id %X is: %s\n", id, nextP->fmtP);
      strncpy(bufP, nextP->fmtP, bufSize);
      return 0;
    }
  }

  /* not found */
  nI2NFail++;
  return -1;
}

void ID2Name(FILE* fmt_fp,trc_datahdr_t *recP,i2nSpec_t* i2nP, idname_t* id2nameP,argType *argP)
{
  int i,j,id;
  Boolean msgHandler;
  /* Update the argP[] array as necessary to do the substitutions */
  if (i2nP != NULL)
  {
    /* This loop proceses each of the substitution strings */
    for (i = 0; i2nP[i].i2nNameP != NULL; i++)
    {
      int whichSub = i2nP[i].argNum - 1;
      char *argtP  = i2nP[i].i2nNameP;
      char *sP = id2nameP[whichSub].s;

      id = argP[whichSub];

      /* Thread_id2name conversion need special treatment */
      msgHandler = false;
      if (strcmp(argtP, "Thread_id2name") == 0)
      {
        /* Thread IDs >= FirstMsgHandlerThread are really message
           handler IDs, so look for a tscMsg_id2name entry in that case.
           Thread IDs < FirstMsgHandlerThread are "standard" thread IDs,
           so look for StdThread_id2name in that case. */
        if (id >= FirstMsgHandlerThread)
        {
          /* Adjust id value and conversion function.  Set a flag to
             remember to insert "Msg handler" in front of the name that
             is found. */
          id -= FirstMsgHandlerThread;
          argtP = "tscMsg_id2name";
          msgHandler = true;
        }
        else
          argtP = "StdThread_id2name";
      }

      /* Now look-up the substitution for argtP (xx_id2name) that is
         associated with the appropriate traced value (id). */
      if (0 != GetID2NameRecord(fmt_fp, argtP, id, TRC_MAX_ID2NAME, sP))
      {
        /* Unable to find the formatting (substitution) information */

        /* Special notation for NodeAddrs */
        if (0 == strcmp(argtP, "nodeaddr_id2name"))
        {
          unsigned int cl, idx;
          switch (id)
          {
            case -1: sprintf(id2nameP[whichSub].s, "<none>"); break;
            case -2: sprintf(id2nameP[whichSub].s, "<invalid>"); break;
            case -3: sprintf(id2nameP[whichSub].s, "<loopback>"); break;
            default:
              cl = (id >> 24) & 0xff;
              idx = id & 0xffff;
              if (id & 0x10000)
                sprintf(id2nameP[whichSub].s, "<c%dp%d>", cl, idx);
              else
                sprintf(id2nameP[whichSub].s, "<c%dn%d>", cl, idx);
              break;
          }
        }

        /* IpAddrs are traced as four words and printed in either
           IPv4 or IPv6 format. */
        else if (0 == strcmp(argtP, "ipaddr_id2name"))
        {
          union { unsigned int w[4]; unsigned char c[16]; } a;

          a.w[0] = a.w[1] = a.w[2] = a.w[3] = 0;
          for (j = 0;
               j < 4 && whichSub + j < LXTRACE_MAX_FORMAT_SUBS;
               j++)
            a.w[j] = argP[whichSub + j];
          *sP = '\0';

          if (a.w[0] == -1U && a.w[1] == -1U &&
              a.w[2] == -1U && a.w[3] == -1U)
            sprintf(sP, "<none>");
          else if (a.w[0] == 0 && a.w[1] == 0 &&
                   a.c[8] == 0 && a.c[9] == 0 &&
                   a.c[10] == 0xff && a.c[11] == 0xff)
            /* Treat IPv4-mapped IPv6 addresses as IPv4 */
            inet_ntop(AF_INET, &a.c[12], sP, TRC_MAX_ID2NAME);
#ifdef AF_INET6
          //not supported on windows yet
          else
            inet_ntop(AF_INET6, a.c, sP, TRC_MAX_ID2NAME);
#endif
          /* Shift following arguments left */
          for (j = whichSub + 1;
               j < recP->trNArgs && j + 3 < LXTRACE_MAX_FORMAT_SUBS;
               j++)
              argP[j] = argP[j + 3];
        }

        /*  ObjKey is traced as three 64-bit hex numbers */
        else if (0 == strcmp(argtP, "objkey_id2name"))
        {
#if defined( __64BIT__) || defined(_LP64)
          long long *w = argP + whichSub;
          const int shift = 2;
#else
          long long w[3];
          const int shift = 5;
          argType *aP = argP + whichSub;
          w[0] = ((unsigned long long)aP[0] << 32) + (unsigned long)aP[1];
          w[1] = ((unsigned long long)aP[2] << 32) + (unsigned long)aP[3];
          w[2] = ((unsigned long long)aP[4] << 32) + (unsigned long)aP[5];
#endif
          if (w[2] == 0xFFFFFFFFFFFFFFFFLL)
            sprintf(sP, "%08X:%08X:%08X:%08X",
                    High32(w[0]), Low32(w[0]), High32(w[1]), Low32(w[1]));
          else
            sprintf(sP, "%016llX:%016llX:%016llX", w[0], w[1], w[2]);

          /* Shift following arguments left */
          for (j = whichSub + 1;
               j < recP->trNArgs && j + shift < LXTRACE_MAX_FORMAT_SUBS;
               j++)
            argP[j] = argP[j + shift];
        }

        /* lockState to string conversions. */
        else if ((0 == strcmp(argtP, "lockstateHigh_id2name")))
        {
          #define lsAcqPend  0x80000000
          #define lsLxLock   0x40000000
          #define lsXwLock   0x20000000
          #define lsWaLock   0x10000000
          #define lsWwLock   0x08000000
          #define lsRvkLock  0x04000000
          #define lsRoCount  0x003FF800
          #define lsRoUnit   0x00000800
          #define lsWoCount  0x000007FF
          #define lsWoUnit   0x00000001
          char buf1[32], buf2[32];
          buf1[0]='\0';
          buf2[0]='\0';
          if (id & lsRoCount)
            sprintf(buf1, " ro:%d", (id & lsRoCount)/lsRoUnit);
          if (id & lsWoCount)
            sprintf(buf2, " wo:%d", (id & lsWoCount)/lsWoUnit);
          sprintf(sP, "[%s%s%s%s%s%s%s%s",
                  (id & lsAcqPend) ? " acq" : "",
                  (id & lsLxLock) ? " lx" : "",
                  (id & lsXwLock) ? " xw" : "",
                  (id & lsWaLock) ? " wa" : "",
                  (id & lsWwLock) ? " ww" : "",
                  (id & lsRvkLock) ? " rvk" : "",
                  buf1, buf2);

        }

        /* lockState to string conversions. */
        else if ((0 == strcmp(argtP, "lockstateLow_id2name")))
        {
          #define lsRfCount  0x7FF00000
          #define lsRfUnit   0x00100000
          #define lsWfCount  0x000FFE00
          #define lsWfUnit   0x00000200
          #define lsRsCount  0x000001FF
          #define lsRsUnit   0x00000001
          char buf1[32], buf2[32], buf3[32];
          buf1[0]='\0';
          buf2[0]='\0';
          buf3[0]='\0';
          if (id & lsRfCount)
            sprintf(buf1, " rf:%d", (id & lsRfCount)/lsRfUnit);
          if (id & lsWfCount)
            sprintf(buf2, " wf:%d", (id & lsWfCount)/lsWfUnit);
          if (id & lsRsCount)
            sprintf(buf3, " rs:%d", (id & lsRsCount)/lsRsUnit);
          sprintf(sP, "%s%s%s ]", buf1, buf2, buf3);
        }

        /* lockFlags to string conversions. */
        else if ((0 == strcmp(argtP, "lockflags_id2name")))
        {
          #define lfDaemonWriter      0x0001
          #define lfRevokePending     0x0002
          #define lfRevokeProcessing  0x0004
          #define lfAcqGranted        0x0008
          sprintf(sP, "[%s%s%s%s ]",
                  (id & lfDaemonWriter) ? " dmn" : "",
                  (id & lfRevokePending) ? " rvk" : "",
                  (id & lfRevokeProcessing) ? " pro" : "",
                  (id & lfAcqGranted) ? " gra" : "");
        }

        /* Any other case is a lack of proper formatting information
           for the generated trace record.  Insert "unknown" into the
           formatted output and continue. */
        else
        {
#ifdef DBGASSERTS
          DP("%s: %s substitution %d (value=%d) not found (used in hook %08X).\n",
             TRCFMT_NAME, argtP, whichSub+1, argP[whichSub], (recP->trHook&0xFFFFFFFF));
          // To debug one of these, exit here.
          // goto exit_format;
#endif // DBGASSERTS
          sprintf(sP, "unknown(%d)", argP[whichSub]);
        }
      }
      
      /* If we are converting the thread ID of a message handler,
         insert "Msg handler" in front of the name. */
      if (msgHandler)
      {
        char tmpBuff[TRC_MAX_ID2NAME];
        strncpy(tmpBuff, sP, sizeof(tmpBuff));
        snprintf(sP, TRC_MAX_ID2NAME, "Msg handler %s", tmpBuff);
      }

      /* Alter the TRC_ARGLIST for this substituted parameter */
      argP[whichSub] = (argType)sP;
    } /* for each substitution string */

  } /* End-substitutions */

}
#ifndef GPFS_WINDOWS
/* Convert timestamp to string format. */
extern char *ctime_r(const time_t *, char *);
char *fmtTime(time_t time, char *timeStr)
{
  /* ctime_r returns a null-terminated, readable string that includes
   * a newline char which we remove.  When ctime_r cannot convert the
   * supplied timestamp, it returns NULL.
   */
  if (!ctime_r(&time, timeStr))
    // append space for formatting
    return "unknown                 ";

  timeStr[strlen(timeStr) - 1] = '\0';
  return timeStr;
}

static struct timeval wrap_prev = { 0, 0 };
static struct timeval wrap_next = { 0, 0 };

/* isWrapPoint:
 * See if consecutive timestamps are backwards (indicates a wrap-point).
 */
int isWrapPoint(struct timeval prev, struct timeval next)
{
  /* Detect timestamps going backwards.  This could be the wrap-point
   * or just wobble.
   */
  if ((prev.tv_sec > next.tv_sec) ||
      ((prev.tv_sec == next.tv_sec)
        && (prev.tv_usec - next.tv_usec > allowed_usec_wobble)))
  {
    /* Prepare for simple subtraction (do any necessary borrowing) */
    if (prev.tv_usec < next.tv_usec)
    {
      prev.tv_usec += 1000000;
      prev.tv_sec  -= 1;
    }

    DFP("isWrapPoint: wrap delta %d.%06d (prev %d.06d, next %d.06d)\n",
        (prev.tv_sec - next.tv_sec), (prev.tv_usec - next.tv_usec),
        prev.tv_sec, prev.tv_usec, next.tv_sec, next.tv_usec);

    /* Detect and warn if multiple wrap points are detected. */
    if (((wrap_prev.tv_sec != 0) &&
         ((wrap_prev.tv_sec != prev.tv_sec) || (wrap_prev.tv_usec != prev.tv_usec))) ||
        ((wrap_next.tv_sec != 0) &&
         ((wrap_next.tv_sec != next.tv_sec) || (wrap_next.tv_usec != next.tv_usec))))
    {
      DP("%s", "Multiple wrap-points found:\n");
      DP("\t%d.%06d-%d.%06d (delta %d.%06d)\n",
         wrap_prev.tv_sec, wrap_prev.tv_usec, wrap_next.tv_sec, wrap_next.tv_usec,
         (wrap_prev.tv_sec - wrap_next.tv_sec), (wrap_prev.tv_usec - wrap_next.tv_usec));
      DP("\t%d.%06d-%d.%06d (delta %d.%06d)\n",
         prev.tv_sec, prev.tv_usec, next.tv_sec, next.tv_usec,
         (prev.tv_sec - next.tv_sec), (prev.tv_usec - next.tv_usec));
    }

    /* Keep track of what we have now declared to be a wrap point */
    wrap_prev = prev;
    wrap_next = next;

    return true;
  }

  return false;
}

/* Read the next trace header from the raw trace file.  Normally, this just
   involves the read() for the specified file.  However, it also needs to
   check for wrapping of the file and make the necessary adjustments when we
   reach the wrap point. */
int ReadNextTrcHeader(FILE *raw_fp, trc_hdrs_t* hdrP, int cpu,
                      long long prevSeq[])
{
  int nRead;
  off_t off1, off2;
  off1 = off2 = 0;

  if (0 < (nRead = fread(hdrP, 1, hdrSize[cpu], raw_fp)))
  {
    /* Successful read.  Validate the header. */
    if (hdrP->hdr.trMagic != LXTRACE_MAGIC &&
        hdrP->hdr.trMagic != LXTRACE_SEQ_MAGIC)
    {
      /* Maybe we've reached the wrap point.  See if we can get oriented.
         Backup and try 1 byte further until we find a valid header.  This
         should be the wrap point.  If we can verify this, return the header.
       */
      fseek(raw_fp, 1-nRead, SEEK_CUR);
      off1 = ftell(raw_fp) - 1;
      while (0 < (nRead = fread(hdrP, 1, hdrSize[cpu], raw_fp)))
      {
        /* Keep inching forward looking for a valid device header that
         * includes timestamps that also indicate a wrap ocurred. */
        if (hdrP->hdr.trMagic == LXTRACE_SEQ_MAGIC && prevSeq[cpu] > hdrP->hdr.trSeq || 
            (hdrP->hdr.trMagic == LXTRACE_MAGIC && isWrapPoint(prevTime[cpu], hdrP->hdr.trTime))) 
        {
          DFP("ReadNextTrcHeader: wrap point between %d.%d %llu and %d.%d %llu\n",
              hdrP->hdr.trTime.tv_sec, hdrP->hdr.trTime.tv_usec, prevSeq[cpu],
              prevTime[cpu].tv_sec, prevTime[cpu].tv_usec, hdrP->hdr.trSeq);

          prevSeq[cpu] = hdrP->hdr.trSeq;
          /* We've found what appears to be a valid header. */
          prevTime[cpu] = hdrP->hdr.trTime; /* timestamp of the last valid record */
          nWraps += 1;
          DFP("ReadNextTrcHeader: wrap point at %d gap %d\n", off1, off2-off1);

          /* Save the oldest timestamp that we've found */
          if (hdrP->hdr.trTime.tv_sec < startTime)
            startTime = hdrP->hdr.trTime.tv_sec;

          return nRead;
        }
        else if (hdrP->hdr.trMagic == LXTRACE_MAGIC ||
                 hdrP->hdr.trMagic == LXTRACE_SEQ_MAGIC)
        {
          /* We've found a valid header some ways from the beginning of
             the file, but it's not a wrap point.  This could be due to
             some garbage in the file.  Try to skip garbage data. */
          prevSeq[cpu] =hdrP->hdr.trSeq;
          /* timestamp of the last valid record */
          prevTime[cpu] = hdrP->hdr.trTime;

          /* Save the latest timestamp we've found */
          if (hdrP->hdr.trTime.tv_sec > stopTime)
            stopTime = hdrP->hdr.trTime.tv_sec;

          /* Save the oldest timestamp that we've found */
          if (hdrP->hdr.trTime.tv_sec < startTime)
            startTime = hdrP->hdr.trTime.tv_sec;

          return nRead;
        }
        fseek(raw_fp, 1-nRead, SEEK_CUR);
        off2 = ftell(raw_fp);
      }
      DP("%s", "lxtrace format: ReadNextTrcHeader failed to find a valid record!\n");
    }
    else
    {
      /* Valid header found, but this could still be a wrap-point if the
       * alignment worked out exactly.
       */
        if (hdrP->hdr.trMagic == LXTRACE_SEQ_MAGIC && prevSeq[cpu] > hdrP->hdr.trSeq || 
            (hdrP->hdr.trMagic == LXTRACE_MAGIC && isWrapPoint(prevTime[cpu], hdrP->hdr.trTime))) 
      {
        DFP("ReadNextTrcHeader: wrap point between %d.%d %llu and %d.%d %llu\n",
            hdrP->hdr.trTime.tv_sec, hdrP->hdr.trTime.tv_usec, prevSeq[cpu],
            prevTime[cpu].tv_sec, prevTime[cpu].tv_usec, hdrP->hdr.trSeq);
        
        nWraps += 1;
        DFP("ReadNextTrcHeader: wrap point at %d gap %d\n", off1, off2-off1);
        
        prevSeq[cpu] = hdrP->hdr.trSeq;
        /* timestamp of the last valid record */
        prevTime[cpu] = hdrP->hdr.trTime;

        /* Save the oldest timestamp that we've found */
        if (hdrP->hdr.trTime.tv_sec < startTime)
          startTime = hdrP->hdr.trTime.tv_sec;
      }
      else
      {
        prevSeq[cpu] = hdrP->hdr.trSeq;
        /* timestamp of the last valid record */
        prevTime[cpu] = hdrP->hdr.trTime;

        /* Save the latest timestamp we've found */
        if (hdrP->hdr.trTime.tv_sec > stopTime)
          stopTime = hdrP->hdr.trTime.tv_sec;

        /* Save the oldest timestamp that we've found */
        if (hdrP->hdr.trTime.tv_sec < startTime)
          startTime = hdrP->hdr.trTime.tv_sec;
      }
    }

  }
  return nRead;
}

/* Read the next trace header from the raw trace file.  Normally, this just
   involves the read() for the specified file.  However, it also needs to
   check for wrapping of the file and make the necessary adjustments when we
   reach the wrap point. */
int ReadNextCompHeader(FILE *raw_fp, trc_comp_header_t *hdrP)
{
  int nRead;

  if (0 < (nRead = fread(hdrP, 1, sizeof(trc_comp_header_t), raw_fp)))
  {
    /* Successful read.  Validate the header. */
    if (hdrP->trMagic != LXTRACE_COMP_MAGIC)
    {
      DFP("possible wrap point at %d\n", ftell(raw_fp)-sizeof(trc_comp_header_t));
      /* Maybe we've reached the wrap point.  See if we can get oriented.
         Backup and try 1 byte further until we find a valid header.  This
         should be the wrap point.  If we can verify this, return the header.
       */
      fseek(raw_fp, 1-nRead, SEEK_CUR);
      while (0 < (nRead = fread(hdrP, 1, sizeof(trc_comp_header_t), raw_fp)))
      {
        if (hdrP->trMagic == LXTRACE_COMP_MAGIC )
        {
          DFP("wrap point found at %d\n", ftell(raw_fp)-sizeof(trc_comp_header_t));
          nWraps++;
          return nRead;
        }
        fseek(raw_fp, 1-nRead, SEEK_CUR);
      }
    }
  }
  return nRead;
}
FILE* findWrapPoint(char *rawPrefix, char *rawSuffix, int cpu, 
                    int nTrcHdrs[], Boolean hasWrapped[], 
                    long long prevSeq[])
{
  char rawTraceFileNameP[PATH_MAX];
  trc_hdrs_t h;
  FILE* raw_fp = NULL;
  int ret;

  if (rawPrefix != NULL)
    {
      if (rawSuffix != NULL)
        sprintf(rawTraceFileNameP, "%s.%s.cpu%d", rawPrefix, rawSuffix, cpu);
      else
        sprintf(rawTraceFileNameP, "%s.cpu%d", rawPrefix, rawSuffix, cpu);

    }
  else
    return NULL;
  /* Open the trace file. */
  if (NULL == (raw_fp = fopen(rawTraceFileNameP, "r")))
  {
    return NULL;
  }

  hasWrapped[cpu] = false; /* Signal format processing below we have a wrap file */
  nWraps = 0;         /* ReadNextTrcHeader counts wrap points it encounters.*/

  nTrcHdrs[cpu] = 0;
  while (0 < ReadNextTrcHeader(raw_fp, &h, cpu, prevSeq))
  {
    nTrcHdrs[cpu]++;

    /* If ReadNextTrcHeader detects a wrap, leave and begin formating. */
    if (nWraps)
      break;

    /* skip over trace data */
    fseek(raw_fp, h.hdr.trLength - sizeof(trc_datahdr_t), SEEK_CUR);
  }

  /* If a wrap point was located, back-up so the header can be read again
     as formatting begins with this record.  If the file never wrapped,
     reset to the top of the file and process top-to-bottom. */
  if (nWraps)
  {
    fseek(raw_fp, -hdrSize[cpu], SEEK_CUR);
    hasWrapped[cpu] = true; /* Signal format procesing below to handle the wrap */
  }
  else
    rewind(raw_fp);
  return raw_fp;
}

/* Defect 816587. lxtrace wants a prefix passed in,
   and lxtrace will append the cpuX suffix
   if you pass in the full name of a file, make sure that
   it doesn't end in "cpuX".
   This is done as a preprocessing optimization to prevent
   unnecessary renaming of trcfiles
*/
void checkFileNamePrefix(char *rawPrefix)
{
  DFP("checking the file name prefix for %s\n",rawPrefix);
  struct stat buf;
  int dateStamp, hour, minute, second, noWrapNum;
  char hostName[PATH_MAX];  
  int strlength = strlen(rawPrefix);
  int i = strlength;

  int rc = gethostname(hostName, PATH_MAX);
  if (rc != 0)
    return;

  if (stat(rawPrefix, &buf) != 0) //if the file doesn't exist, just return
    return;
  if (strstr(hostName, "cpu")) //if the hostName has "cpu" in it, return
    return;

  //now check if the end of the string is "cpu%d"
  for (i = strlength -1; rawPrefix[i] != '.'; i--)
  {
    if (!isdigit(rawPrefix[i]))
      break;
    if (i == 0)
      break;
  }
  if (i < 3) //If we might look before the start of the string, abort now
    return;
  if (rawPrefix[i] == 'u' && rawPrefix[i-1] == 'p' && 
      rawPrefix[i-2] == 'c' && rawPrefix[i-3] == '.')
    rawPrefix[i-3] = '\0';
}

//find raw files
int findRawFiles(char *rawPrefix)
{
  char fileName[PATH_MAX];
  int nCPU = 0;
  int rc;
  struct stat buf;

  if(stat(rawPrefix,&buf) == 0)
  {
    if(buf.st_size <= 0)
      return 0;
    sprintf(fileName, "%s.cpu0", rawPrefix);
    rename(rawPrefix, fileName);
    return 1;
  }
 
  while (1)
  {
    sprintf(fileName, "%s.cpu%d", rawPrefix, nCPU);
    if(stat((const char*)fileName,&buf) != 0)
      break;
    nCPU++;
  }
  
  return nCPU;
}

//uncompress raw trace file
FILE *unComp(char *compPrefix, char *compSuffix, int cpu,
             Boolean hasWrapped[])
{
  char compTraceFileNameP[PATH_MAX];
  char uncompFileNameP[PATH_MAX];
  FILE *comp_fp; /* raw trace data file */
  FILE *out_fp; /* output file */
  trc_comp_header_t cmp_hdr;
  unsigned long dlen;
  Boolean compressed;
  void *libz_handle = NULL;
  char *error;

  comp_fp = out_fp = NULL;
  /* work buffers */
  char *src = NULL;   /* compressed  trace data */
  char *dst = NULL;   /* uncompressed trace data */
  int ret;

  sprintf(compTraceFileNameP, "%s.cpu%d", compPrefix, cpu);
  /* Open the trace file. */
  if (NULL == (comp_fp = fopen(compTraceFileNameP, "r")))
  {
    goto exit_uncomp;
  }

  if ( fread(&cmp_hdr, 1, sizeof(trc_comp_header_t), comp_fp) > 0)
  {
    //search wrap point
    if (cmp_hdr.trMagic == LXTRACE_COMP_MAGIC)
    {
      hdrSize[cpu] = sizeof(trc_hdrs_t);
      fseek(comp_fp, 0, SEEK_SET);
      src = malloc(cmp_hdr.dataLength);
      if ( src == NULL )
        goto exit_uncomp;
      dst = malloc(cmp_hdr.dataLength);
      if ( dst == NULL )
        goto exit_uncomp;
      nWraps = 0;
      //read compression record until wrap point found or reach file tail
      while (ReadNextCompHeader(comp_fp, &cmp_hdr)>0)
      {
        DFP("unComp: find record %d at %d\n",
            cmp_hdr.compLength, ftell(comp_fp));
        if (nWraps)
        {
          break;
        }
        fseek(comp_fp, cmp_hdr.compLength, SEEK_CUR);
      }
      //no wrap point found, go back to head
      if (!nWraps)
        rewind(comp_fp);
      //find wrap point, seek to it
      else
      {
        fseek(comp_fp, -sizeof(trc_comp_header_t), SEEK_CUR);
        hasWrapped[cpu] = true;
      }
    }
    else
    {
      //for backward compatibility
      //previous trace header do not have sequence number
      //Also, non-blocking (LXTRACE_NB) traces aren't compressed
      if (cmp_hdr.trMagic == LXTRACE_MAGIC)
        hdrSize[cpu] = sizeof(trc_noseq_header_t) + sizeof(trc_datahdr_t);
      else if (cmp_hdr.trMagic == LXTRACE_SEQ_MAGIC)
        hdrSize[cpu] = sizeof(trc_hdrs_t);
      DFP("unComp: header magic %x header size %d\n",
          cmp_hdr.trMagic, hdrSize[cpu]);
      goto exit_uncomp;
    }

    //open uncompressed raw trace file
    sprintf(uncompFileNameP, "%s.%s.cpu%d", compPrefix, compSuffix, cpu);


    out_fp = fopen(uncompFileNameP, "w+");
    if (out_fp == NULL)
    {
      DP("lxtrace format: could not open file %s, errno %d\n",
         uncompFileNameP, errno);
      goto exit_uncomp;
    }
    //load libz.so.1
    libz_handle = dlopen("libz.so.1",RTLD_NOW);
    if( libz_handle == NULL )
    {
      DP("lxtrace format: dlopen libz.so.1 error: %s\n",dlerror());
      goto exit_uncomp;
    }
    //clear previous dlerror
    dlerror();
    uncompress = dlsym(libz_handle, "uncompress"); 
    if ((error = dlerror()) != NULL) 
    {
      DP("lxtrace format: dlsym uncompress error: %s\n",error);
      goto exit_uncomp;
    } 
    //start uncompression from wrap point
    nWraps = 0;
    while (1)
    {
      if (ReadNextCompHeader(comp_fp, &cmp_hdr)>0)
      {
        if (nWraps)
          break;
        ret = fread(src, 1, cmp_hdr.compLength, comp_fp);
        if (ret == cmp_hdr.compLength)
        {
          dlen = cmp_hdr.dataLength;
          ret = uncompress(dst, &dlen, src, cmp_hdr.compLength);
          if ( ret != 0 )
          {
            if( ret == -3 )
              DP("%s", "lxtrace format: uncompression data error\n");
            else if( ret == -5 )
              DP("%s", "lxtrace format: uncompression buffer error\n");
            else
              DP("%s", "lxtrace format: uncompression error %d\n",ret);
          }
          else if ( dlen != 0 )
          {
            totalCompSize += cmp_hdr.compLength;
            totalPostCompSize += dlen;
            ret = fwrite(dst, dlen, 1, out_fp);
            if(ret == 0)
            {
              DP("lxtrace format: fwrite error %s\n",strerror(errno));
              goto exit_uncomp;
            }
          }
        }
        else
          DP("lxtrace format: compressed file incomplete. Sought %d bytes, get %d bytes\n", cmp_hdr.compLength, ret);
      }
      //go back to file head and uncompress
      else if (hasWrapped[cpu])
      {
        rewind(comp_fp);
        hasWrapped[cpu] = false;
      }
      else break;
    }
  }

exit_uncomp:
  if(libz_handle != NULL)
    dlclose(libz_handle);
  if (src != NULL)
    free(src);
  if (dst != NULL)
    free(dst);
  if (comp_fp != NULL) fclose(comp_fp);
  if (out_fp != NULL)
  {
    rewind(out_fp);
  }
  return out_fp;
}
//pick up next trace record
trc_hdrs_t* PickNextTrcHeader(FILE* raw_fp[], int nCPU, trc_hdrs_t cpuHdr[],
                              Boolean hasWrapped[], long long prevSeq[])
{
  int i;
  //one trace file
  if (nCPU == 1)
  {
    if (ReadNextTrcHeader(raw_fp[0], &cpuHdr[0], 0, prevSeq)>0)
    {
      if (nWraps ) // ReadNextTrcHeader detected and processed a wrap-point
      {
        DFP("%s", "lxtrace format: ending after processing records 0 to wrap-point\n");
        nWraps = 0;
        return NULL;
      }
      return &cpuHdr[0];
    }
    else
      return NULL;
  }
#if defined(GPFS_LINUX) && defined(LXTRACE_USE_RELAY)
  //this cpu raw file still have trace data
  if (cpuHdr[curCPU].hdr.trMagic != LXTRACE_NODATA_MAGIC)
  {
    //read trace data from last accessed cpu file
    if (ReadNextTrcHeader(raw_fp[curCPU], &cpuHdr[curCPU], curCPU, prevSeq)>0)
    {
      /* Since we always start AFTER the wrap point, we should only see a
         wrap if we had reached the end of the file and then gone back to
         the beginning to get the records from there up to the wrap point.
         So, if we see this, we have formatted all the records and are done. */

      if (nWraps ) // ReadNextTrcHeader detected and processed a wrap-point
      {
        DFP("%s", "lxtrace format: ending after processing records 0 to wrap-point\n");
        cpuHdr[curCPU].hdr.trMagic = LXTRACE_NODATA_MAGIC;
        nWraps = 0;
      }
    }
    //go back to caller for rewind
    else if (hasWrapped[curCPU])
      return NULL;
    //no data left
    else
      cpuHdr[curCPU].hdr.trMagic = LXTRACE_NODATA_MAGIC;

  }

  //find next cpu file which has data to read
  curCPU = nCPU;
  for (i = 0; i < nCPU; i++)
  {
    if (cpuHdr[i].hdr.trMagic != LXTRACE_NODATA_MAGIC)
    {
      curCPU = i;
      break;
    }
  }
  //all cpu files are consumed
  if (curCPU>=nCPU) return NULL;

  //only last cpu file has data
  if (curCPU == nCPU - 1)
    return &cpuHdr[curCPU];

  //pick up oldest trace record from each cpu trace file
  for (i=curCPU+1; i<nCPU; i++)
  {
    if (cpuHdr[i].hdr.trMagic != LXTRACE_NODATA_MAGIC)
    {
      if(cpuHdr[i].hdr.trSeq < cpuHdr[curCPU].hdr.trSeq) 
      {
        curCPU = i;
      }
    }
  }
#endif
  return &cpuHdr[curCPU];
}

/* ProcessBuffer reads from the trace device and writes to the output file.
   This is a function used by the trace daemon when handling SIGIO from the
   device or after fsync() as the device is being closed. */
void ProcessBuffer()
{
  int rc, nBytes;

retry:

  /* Issue a read for one buffer to the trace device */
  DFP("trace daemon about to read %d bytes\n", bufSize);
  nBytes = read(TraceHandle, bufP, bufSize);
  DFP("trace daemon read %d bytes\n", nBytes);
  if (nBytes == -1)
  {
    DP("trace daemon read error %d\n", errno);
    close(TraceHandle);
    closeOutFiles(0);
    _exit(1);
  }
  if (nBytes > 0)
  {
    compWrite(0, bufP, nBytes);
    goto retry;
  }

}

/* The daemon signal handler (when not using LXTRACE_USE_RELAY).
   This is the main part of the daemon which (after initializing)
   does nothing except wait for signals.  SIGIO indicates that
   a trace buffer is ready to be processed, any other signal stops tracing and
   terminates the daemon. */
void SignalHandler(int signum)
{
  int rc;
  struct kArgs args;
  trc_msg msg;

  DFP("trace daemon got signal %d\n", signum);
  switch (signum)
  {
#if defined(GPFS_LINUX)
    case SIGIO:
#elif defined(GPFS_SOLARIS)
    case SIGURG:
#endif
      /* This is the main daemon function.  Handle SIGIO signals from the
         device and transfer the data from the device to the output file
         when received.  */
      ProcessBuffer();
      break;

    default:
      /* SIGTERM (or other) will terminate the daemon and end the trace. */
      // DP("trace daemon: terminating due to signal %d\n", signum);
      if (TraceHandle > 0)
      {
        /* Get any records that are buffered in the device. */
        ProcessBuffer();
        flushAioBuffers(0);
        msgQueue = msgGet(0);
        if( msgQueue == -1 )
        {
          DP("trace daemon: msgget error %d\n",errno);
        }
        else
        {  
          msg.mtype = LXTRACE_MTYPE;
          msgsnd(msgQueue, &msg, 0, 0);
          DFP("trace daemon: send message to lxtrace off\n");
        }
        freeBuffers();
        if (nWraps)
          DP("lxtrace: Trace file wrapped %d times.\n", nWraps);
        /* close everything */
        close(TraceHandle);
        closeOutFiles(0);
      }

      /* daemon terminates */
      if (bufP)
        free(bufP);

      _exit(0);
  }
}


/* Print command syntax and exit */
void Usage()
{
  DP("%s", "\n\
Usage: lxtrace { on on-args | off | format format-args | dump | fsync }\n\
\n\
  Command line format for lxtrace on:\n\
    lxtrace on [trcFile] [-r] [-p] [-w] [-s filesize] [-b bufsize] [-n bufferNum]\n\
               [-i interval] [-c complevel] [-l logFile]\n\
      -r         Turn on relayfs trace\n\
      -w         Turn on overwrite mode\n\
      -p         Turn on per-cpu buffer trace\n\
      -t         Turn on coarse-grained timestamp\n\
      -k         Rename raw trace file upon file wrap to keep all trace data\n\
      trcFile    Name of raw trace file to be created.  Default is\n\
                   /tmp/lxtrace.trc.\n\
      filesize   Optional maximum size of wrap-around trace file.  A value\n\
                   of 0 means that the trace file grows indefinitely.  Default\n\
                   is 16M.\n\
      bufsize    Optional size of each buffer.  Default is 4M.\n\
      bufferNum  Optional subbuffer number fora each cpu.  Default is 4.\n\
      interval   Optional trace flush interval.  Default is 15.\n\
      complevel  Optional trace compression level.  Default is 6.\n\
      logFile    Optional file to log all lxtrace daemon activity.\n\
                 Default is to stdout/stderr\n\
\n\
  Command line format for lxtrace format:\n\
    lxtrace format [-T | -D] [-v] [-t formatFile] [-w wobble] [-o outFile] [trcFile]\n\
      -T         Display timestamps as seconds since 1/1/1970.  Default\n\
                   is to display the first timestamp as 0.0 and other\n\
                   timestamps as elapsed time since the first trace record.\n\
      -D         Display timestamps in local calendar format. This option \n\
                   can not be used together with -T option.\n\
      -v         Verbose mode.  Include CPU number, buffer number, and hook\n\
                  word in trace output.  Default is to not include these.\n\
      wobble     Number of microseconds wobble allowed in timestamps.\n\
                 An apparent wrap within the wobble limit will be ignored.\n\
      formatFile Name of file containing trace formats.  Default is\n\
                   /usr/lpp/mmfs/mmfs.trcfmt.\n\
      outFile    Name of file to receive formatted trace output.  Default\n\
                   is standard output.\n\
      trcFile    Name of input raw trace file.  Default is /tmp/lxtrace.trc.\n\
      --stat     Generate statistics about the trcfile being formatted\n"
     );
  if (TraceHandle != -1)
    close(TraceHandle);
  exit(1);
}


/* The lxtrace command.  */
int main(int argc, char *argv[])
{
  int rc = 0;
  struct kArgs args; // command arguments
  int i;
  int nTrcHdrs[LXTRACE_NR_CPUS];
  Boolean hasWrapped[LXTRACE_NR_CPUS] = { 0 };
  if (argc < 2)
    Usage();

  /* A new trace is being started.  Validate the options, open the files,
     start the daemon and configure the device. */
  if (strcmp(argv[1], "on") == 0)
  {
    int hw, dpid;
    struct sigaction action;
    char* trcFileNameP = NULL;
    trc_msg msg;
    int mode;
    char *daemonLogFile = NULL;
    /* Initialize local state fields. */
    bufP = NULL;
    outSize = DEF_TRC_FILESIZE;
    bufNum = DEF_TRC_BUFNUM;
    interval = DEF_TRC_INT;
    compLevel = DEF_TRC_COMPLEVEL;
    currentPos = 0;
    nWraps = 0;

    /* Parse arguments */
    for (i = 2; i < argc; i++)
    {
      if (strcmp(argv[i], "-r") == 0)
        useRelay = 1;
      else if (strcmp(argv[i], "-p") == 0)
        relayBuf |= RL_BUF_PERCPU;
      else if (strcmp(argv[i], "-t") == 0)
        relayBuf |= RL_BUF_CTIME;
      else if (strcmp(argv[i], "-w") == 0)
        writeMode = LXTRACE_NB;
      else if (strcmp(argv[i], "-k") == 0)
        noWrap = 1;
      else if (strcmp(argv[i], "-s") == 0  &&  i+1<argc)
      {
        i += 1;
        /* Save the output filesize.  If the size is explicitly given as zero,
           do not wrap. */
        outSize = atoll(argv[i]);
      }
      else if (strcmp(argv[i], "-b") == 0  &&  i+1<argc)
      {
        i += 1;
        /* If specified, set the trace device buffer size (otherwise use default) */
        bufSize = atoi(argv[i]) & 0xfffff000;
      }
      else if (strcmp(argv[i], "-n") == 0  &&  i+1<argc)
      {
        i += 1;
        bufNum = atoi(argv[i]);
        if (bufNum < MIN_TRC_BUFNUM || bufNum > MAX_TRC_BUFNUM)
        {
          DP("Buffer number must be between %d and %d\n",
             MIN_TRC_BUFNUM, MAX_TRC_BUFNUM);
          Usage();
        }
      }
      else if (strcmp(argv[i], "-i") == 0  &&  i+1<argc)
      {
        i += 1;
        interval = atoi(argv[i]);
        if (interval < -1)
        {
          DP("Interval must be more than %d\n", -1);
          Usage();
        }
      }
      else if (strcmp(argv[i], "-c") == 0  &&  i+1<argc)
      {
        i += 1;
        compLevel = atoi(argv[i]);
        if (compLevel < MIN_TRC_COMPLEVEL || compLevel > MAX_TRC_COMPLEVEL)
        {
          DP("Compression level must be between %d and %d\n",
             MIN_TRC_COMPLEVEL, MAX_TRC_COMPLEVEL);
          Usage();
        }
      }
      else if (strcmp(argv[i], "-l") == 0  &&  i+1<argc)
      {
        i += 1;
        daemonLogFile = strdup(argv[i]);
        if (daemonLogFile == NULL)
        {
          DP("E_NOMEM in strdup for -l\n");
          exit(1);
        }
      }

      else
        trcFileNameP = argv[i];
    }

    /* Get output file name */
    if( trcFileNameP == NULL )
      trcFileNameP = "/tmp/lxtrace.trc";
    strcpy(percpuOutBasename, trcFileNameP);

    //set default buffer size
    if (!bufSize)
    {
      if ( writeMode == LXTRACE_NB )
        bufSize = DEF_TRC_OVERWRITE_BUFSIZE & 0xfffff000;
      else
        bufSize = DEF_TRC_BUFSIZE & 0xfffff000;
    }

    //ignore trace file size limit for overwrite mode
    if ( writeMode == LXTRACE_NB )
      outSize = 0;

    /* The output filesize must be at least as big as the selected buffer size. */
    if (((outSize < MIN_TRC_FILESIZE) || (outSize > MAX_TRC_FILESIZE)) &&
        (outSize != 0))
    {
      DP("Output file size must be 0 or between %d and %lld\n",
         MIN_TRC_FILESIZE, MAX_TRC_FILESIZE);
      Usage();
    }
    if ((outSize != 0) && (outSize < bufSize))
    {
      DP("lxtrace: the filesize (%d) must be at least as large as the bufsize (%d)\n",
         outSize, bufSize);
      Usage();
    }

    if ( writeMode == LXTRACE_NB )
    {
      // no compression in overwrite mode
      compLevel = 0;
      if (bufSize < MIN_TRC_OVERWRITE_BUFSIZE)
      {
        DP("Overwrite buffer size must be more than %d\n",
           MIN_TRC_OVERWRITE_BUFSIZE);
        Usage();
      }
    }
    else
    {
      if (bufSize < MIN_TRC_BUFSIZE)
      {
        DP("lxtrace: trace buffer has size %d instead of the requested %d\n",
           MIN_TRC_BUFSIZE, bufSize);
        bufSize = MIN_TRC_BUFSIZE;
      }
      else if (bufSize > MAX_TRC_BUFSIZE)
      {
        DP("lxtrace: trace buffer has size %d instead of the requested %d\n",
           MAX_TRC_BUFSIZE, bufSize);
        bufSize = MAX_TRC_BUFSIZE;
      }
    }

    //for recycle and daemon error message
    msgQueue = msgGet(IPC_CREAT | 0666);
    if( msgQueue == -1 )
    {
      DP("trace daemon: msgget error %d\n",errno);
      rc = errno;
      goto exit;
    }  

    /* The trace daemon does all the work.  It is created here ("lxtrace on") and
       lives until "trace off" sends it a SIGTERM signal. */
    dpid = fork();

    if (dpid < 0)
    {
      DP("%s", "lxtrace on: unable to fork trace daemon");
      rc = EAGAIN;
      goto exit;
    }

    else if (dpid != 0)
    {
      /* Parent */
      //wait for child daemon initialization message
      for(i=0;i<100;i++)
      {
        //do not wait in case the child exit abnormally and the parent will be blocked forever   
        rc = msgrcv(msgQueue, &msg, LXTRACE_MSG_LEN, LXTRACE_MTYPE, IPC_NOWAIT); 
        if(rc == -1)
        {
          rc = errno;
          //no message, sleep for 1/10th second and retry
          if(errno == ENOMSG )
          {
            usleep(100000);
          }
          else
          {
            DP("lxtrace on: msgrcv error %d\n",errno);
            break;
          }
        }
        else 
        {
          rc = 0;
          break;
        }
      }
      DFP("lxtrace on: receive message from daemon: %s\n",msg.mtext);
      msgctl(msgQueue,IPC_RMID,NULL);
      //return initialization result to mmtrace
      if(rc != 0 || strcmp(msg.mtext,LXTRACE_INIT_ERR)==0)
      {
        return -1;
      }
      else
        return 0;
    }

    else
    { /* Child/Daemon */

      if (useRelay)
      {
        if(daemonLogFile != NULL)
          {
            int fd = open(daemonLogFile, O_CREAT|O_RDWR|O_SYNC|O_TRUNC, 0644);
            if(fd <= 0)
              {
                //even if the daemonLogFile code fails, still run RelayStart
                fprintf(stdout, "open(%s) failed, rc %d errno %d (%s)\n",
                        daemonLogFile, fd, errno, strerror(errno));
              }
            else
              {
                DFP("open(%s) for daemon log succeeded, using fd %d\n",daemonLogFile,fd);
                int rc = dup2(fd, 1); //reroute stdout to fd
                if(rc != 0)
                  DFP("dup2 failed with rc %d, errno %d (%s\n",
                      rc, errno, strerror(errno));
                rc = dup2(fd, 2); // reroute stderr to fd
                if(rc != 0)
                  DFP("dup2 failed with rc %d, errno %d (%s\n",
                      rc, errno, strerror(errno));
              }
          }

#if defined(GPFS_LINUX) && defined(LXTRACE_USE_RELAY)
        rc = RelayStart(outSize, bufSize, bufNum, interval, compLevel);
#endif
        goto exit_on;
      }

#ifdef LXTRACE_USE_DIO
      //only use DIO for blocking mode
      if ( writeMode == LXTRACE_BLK )
        useDIO=1;
#endif

      if(openOutFiles(0, trcFileNameP)<0)
      {
        rc = errno;
        goto exit_on;
      }
      /* At this point, we are in the trace daemon.  Complete device
         configuration and then wait on its signals. */

      bufP = (char *)malloc(bufSize);
      if (!bufP)
      {
        DP("%s", "trace daemon: unable to malloc trace buffer");
        rc = ENOMEM;
        goto exit_on;
      }
      memset(bufP, 0, bufSize);     /* Force page faults to occur now */
      cpuNum = 1;
      subBufSize = bufSize;
      percpuOutSize = outSize;
      if(allocBuffers()<0)
      { 
        DP("%s", "trace daemon: unable to malloc trace buffer\n");
        rc = ENOMEM;
        goto exit_on;
      }
      /* Initialize the signal handler */
      action.sa_handler = SignalHandler;
      sigfillset(&action.sa_mask);  /* block all signals while the handler is running */
      action.sa_flags = 0;

#if defined(GPFS_LINUX)
      if ((sigaction(SIGIO, &action, NULL)) ||
          (sigaction(SIGTERM, &action, NULL)))
#elif defined(GPFS_SOLARIS)
        if ((sigaction(SIGURG, &action, NULL)) ||
            (sigaction(SIGQUIT, &action, NULL)))
#endif
        {
          DP("%s", "trace daemon: unable to create signal handler\n");
          rc = EINVAL;
          goto exit_on;
        }

      /* Initialize the trace device */

      TraceHandle = open(TRC_DEVICE, O_RDWR);
      if (TraceHandle < 0)
      {
        /* Device does not exist, wrong type, or already open, ... */
        DP("trace daemon: device open failed with errno %d\n", errno);
        rc = ENODEV;
        goto exit_on;
      }

      /* Configure trace device buffer size. */
      args.arg1 = bufSize;
      rc = ioctl(TraceHandle, trc_bufSize, &args);
      if (rc < 0)
      {
        DP("trace daemon: device buffer request failed with errno %d\n", errno);
        rc = ENOMEM;
        goto exit_on;
      }

      //initilization succeed, send message to parent
      //should be put before turn on trace, otherwise it may run into signal handler
      //and parent do not get notified immediately
      msg.mtype = LXTRACE_MTYPE;
      strcpy(msg.mtext,LXTRACE_INIT_OK);
      msgsnd(msgQueue, &msg, strlen(msg.mtext)+1, 0);
      DFP("trace daemon: send message to parent: %s\n",msg.mtext);

      args.arg1 = writeMode;
      /* Mark as ready for tracing. */
      rc = ioctl(TraceHandle, trc_begin, &args);
      if (rc < 0)
      {
        DP("trace daemon: device begin failed with errno %d\n", errno);
        //avoid sending message since parent already exited
        rc = 0;
        goto exit_on;
      }
      
      /* wait here forever (until SIGTERM is handled and exits) */
      while (1)
      {
        DFP("%s", "trace daemon about to wait\n");
        pause();
      }
      //_exit(-1);
    } /* child/daemon */

exit_on:
    if(msgQueue != -1 )
    {
      if(rc!=0)
      {
        //initialization failure, send message to parent
        //the parent will remove the message queue
        msg.mtype = LXTRACE_MTYPE;
        strcpy(msg.mtext,LXTRACE_INIT_ERR);
        msgsnd(msgQueue, &msg, strlen(msg.mtext)+1, 0);
        DFP("trace daemon: send message to parent: %s\n",msg.mtext);
      }
    }
    /* exit here for failures during daemon initialization. */
    DFP("%s", "trace daemon exiting\n");
    if (TraceHandle > 0)
      close(TraceHandle);
  }


  /* Format the trace file. */
  else if (strcmp(argv[1], "format") == 0)
  {
    FILE *raw_fp[LXTRACE_NR_CPUS]={ NULL }; /* raw trace data file */
    FILE *fmt_fp; /* the application formatting file */
    FILE *out_fp; /* output file */
    FILE *stat_fp; /* output file for statistics */
    struct timeval tzero = { 0, 0 };
    trc_hdrs_t *h;
    int useTOD = 0;    /* true if -T option given */
    int useLocal = 0;  /* true if -D option given */
    int verbose = 0;   /* true if -v option given */
    int sequence = 0;   /* true if -s option given */
    char* formatFileNameP = "/usr/lpp/mmfs/mmfs.trcfmt";
    char* outputFileNameP = NULL;
    char *statFileNameP = NULL;
    char* rawTraceFileNameP = "/tmp/lxtrace.trc";
    char* compSuffix = "uncomp";
    char *compP, *fmtP;
    unsigned int llmask;
    i2nSpec_t *i2nP;
    int nRecordsProcessed = 0;  /* useful when running under gdb */
    int i, j, id;
    Boolean msgHandler;
    char timeBufP[26]; /* timestamp in string format */
    FILE *version_fp;
    trc_datahdr_t *recP;
    trc_hdrs_t cpuHdr[LXTRACE_NR_CPUS];
    unsigned long long prevSeq[LXTRACE_NR_CPUS];

    /* work buffers */
    char application_data[LXTRACE_MAX_DATA];   /* raw trace data */

    /* Array of id2name substitutions to be made */
    idname_t id2nameP[LXTRACE_MAX_FORMAT_SUBS];
    char fmtString[256];

    /* Array of pointers to the args for the fprintf that formats the output */
    argType argP[LXTRACE_MAX_FORMAT_SUBS];

    int nCPU = 1;
    int fileWrapped = 0;
    fmt_fp = out_fp = stat_fp = NULL;
    
    formatFileP = NULL;
    fmtDescSpaceP = NULL;

    unsigned long long lastSeq = 0;
    unsigned long long lostRecords = 0;
#if (defined(__64BIT__) || defined(_LP64))
    startTime = 9223372036854775807;
#else
    startTime = 2147483647;
#endif
    stopTime = 0;

    /* Parse arguments */
    for (i = 2; i < argc; i++)
    {
      if (strcmp(argv[i], "-T") == 0)
        useTOD = 1;
      if (strcmp(argv[i], "-D") == 0)
        useLocal = 1;
      else if (strcmp(argv[i], "-v") == 0)
        verbose = 1;
      else if (strcmp(argv[i], "-s") == 0)
        sequence = 1;
      else if (strcmp(argv[i], "--stat") == 0)
        genStats = 1;
      else if (strcmp(argv[i], "--statfile") == 0)
        {
          i += 1;
          statFileNameP = argv[i];
        }
      else if (strcmp(argv[i], "-w") == 0  &&  i+1<argc)
      {
        char *endptr;
        i += 1;
        allowed_usec_wobble = strtol(argv[i], &endptr, 10);
        if ((*endptr != 0) || (allowed_usec_wobble < 0))
        {
          DP("%s", "lxtrace format: -w requires an integer argument greater than or equal to zero.\n");
          rc = 12;
          goto exit_format;
        }
      }
      else if (strcmp(argv[i], "-t") == 0  &&  i+1<argc)
      {
        i += 1;
        formatFileNameP = argv[i];
      }
      else if (strcmp(argv[i], "-o") == 0  &&  i+1<argc)
      {
        i += 1;
        outputFileNameP = argv[i];
      }
      else
        rawTraceFileNameP = argv[i];
    }

    /* Open the formatting file. */
    if (NULL == (fmt_fp = fopen(formatFileNameP, "r")))
    {
      DP("lxtrace format: could not open file %s, errno %d\n", formatFileNameP, errno);
      rc = errno;
      goto exit_format;
    }

    if (0 != (rc = BuildFormatTable(fmt_fp)))
    {
      DP("lxtrace format: could not parse file: %s errno %d\n",
         formatFileNameP, errno);
      rc = errno;
      goto exit_format;
    }

    /* Open and truncate the output file. (defaults to stdout) */
    if (outputFileNameP == NULL)
      out_fp = stdout;
    else
    {
      out_fp = fopen(outputFileNameP, "w");
      if (out_fp == NULL)
      {
        DP("lxtrace format: could not open file %s, errno %d\n",
           outputFileNameP, errno);
        rc = errno;
        goto exit_format;
      }
    }
    /* Open and truncate the stat file. (defaults to stdout) */
    if (statFileNameP == NULL)
      stat_fp = stdout;
    else
    {
      stat_fp = fopen(statFileNameP, "w");
      if (stat_fp == NULL)
      {
        DP("lxtrace format: could not open file %s, errno %d\n",
           statFileNameP, errno);
        rc = errno;
        goto exit_format;
      }
    }

    //the code wants this to be a file prefix, not the full filename
    //make sure this is the case
    checkFileNamePrefix(rawTraceFileNameP);

    nCPU = findRawFiles(rawTraceFileNameP);
    if ( nCPU == 0)
    {
      DP("lxtrace format: could not find raw file %s, errno %d\n",
         rawTraceFileNameP, errno);
      rc = errno;
      goto exit_format;
    }

    /* The trace file (raw_fp) is a wrap-around file.  We need to read it to
       see if it ever wrapped.  If it never wrapped, rewind and process it
       top-to-bottom.  If the file did wrap, set the file pointer to the oldest
       record and begin processing from there.  ReadNextTrcHeader detects the
       wrap point.  When formatting of a wrapped file reaches end-of-file, it
       must then go back and read starting at the beginning and stop when
       the wrap point is again reached. */
    for (i=0; i<nCPU; i++)
    {
      //uncompress and find wrap point
      raw_fp[i] = unComp(rawTraceFileNameP, compSuffix, i, hasWrapped);
      if (raw_fp[i] == NULL)
      {
        /* no compression, find wrap point directly */
        raw_fp[i] = findWrapPoint(rawTraceFileNameP, NULL, i, 
                                  nTrcHdrs, hasWrapped, prevSeq);
      }
      else
      {
        /* Compression was used, so we need to scan through for the
           start/end times. During unComp(), the resulting trcfile.uncomp
           files should have had the wrap point placed at the beginning
           of the file. This scan is just to find the timestamps. */
        Boolean hasWrapped_tmp[LXTRACE_NR_CPUS]; //just a placeholder
        (void) findWrapPoint(rawTraceFileNameP, compSuffix, i,
                             nTrcHdrs, hasWrapped_tmp, prevSeq);
      }
      if (raw_fp[i] == NULL)
        goto exit_format;

      /* Show the point we interpreted as the wrap-point */
      if (nWraps)
      {
        DFP("CPU %d Wrap point  : %d.%06d-%d.%06d (delta %d.%06d)\n", i,
            wrap_prev.tv_sec, wrap_prev.tv_usec, wrap_next.tv_sec, wrap_next.tv_usec,
            (wrap_prev.tv_sec-wrap_next.tv_sec), (wrap_prev.tv_usec-wrap_next.tv_usec));
      }
      DFP("wrap analysis: nTrcHdrs %d nWraps %d seekpos %d\n",
          nTrcHdrs[i], nWraps, ftell(raw_fp[i]) - hdrSize[i]);
    }

    /* Begin printing out header information */
    fprintf(out_fp, FMT_TRC_BEGIN, fmtTime(startTime, timeBufP));
    fprintf(out_fp, FMT_TRC_END, fmtTime(stopTime, timeBufP),
            fileWrapped? FILE_WRAP_STR:FILE_NOWRAP_STR);
    /* Achieved compression ratio */
    if (totalCompSize != 0)
      fprintf(out_fp, FMT_COMP_RATIO, totalPostCompSize / totalCompSize);
    else
      fprintf(out_fp, FMT_COMP_RATIO, 0);

    /* Write kernel version to the trace file.  This is the machine
     * where the format was done, but this would normally be the
     * same machine that generated the raw trace as well.
     */
#ifdef GPFS_SOLARIS
    version_fp = fopen("/etc/nodename", "r");
#else
    version_fp = fopen("/proc/version", "r");
#endif
    if (version_fp)
    {
      int verLen;
      const int verStrLen = 512;
      char version_str[verStrLen];

      memset(version_str, 0, verStrLen);
      if (0 < (verLen = fread(version_str, 1, verStrLen, version_fp)))
      {
        version_str[verStrLen-1] = '\0';
        fprintf(out_fp, "Formatted on: %s\n\n", version_str);
      }

      fclose(version_fp);
    }

    /* Reset to avoid reporting a wrap on the first record from the next
     * call to ReadNextTrcHeader.
     */
    for (i=0; i<nCPU; i++)
    {
      prevTime[i].tv_sec = 0;
      prevTime[i].tv_usec = 0;
      prevSeq[i] = 0;
    }
    /* Print first line of header and build trace format string */
    fmtString[0] = '\0';
    if(sequence)
    {
      fprintf(out_fp, H1_SEQ);
      strcat(fmtString, FMT_SEQ);
    }
    if (useTOD)
    {
      fprintf(out_fp, H1_TS_T);
      strcat(fmtString, FMT_TS_T);
    }
    else if (useLocal)
    {
      fprintf(out_fp, H1_TS_D);
      strcat(fmtString, FMT_TS_D);
    }
    else
    {
      fprintf(out_fp, H1_TS);
      strcat(fmtString, FMT_TS);
    }
    fprintf(out_fp, H1_PID);
    strcat(fmtString, FMT_PID);
    if (verbose)
    {
      fprintf(out_fp, H1_P_HW_V);
      strcat(fmtString, FMT_P_HW_V);
    }
    fprintf(out_fp, H1_TAG);
    strcat(fmtString, FMT_TAG);

    /* Print second line of header */
    fprintf(out_fp, H2_SEQ);
    if (useTOD)
      fprintf(out_fp, H2_TS_T);
    else if (useLocal)
      fprintf(out_fp, H2_TS_D);
    else
      fprintf(out_fp, H2_TS);
    fprintf(out_fp, H2_PID);
    if (verbose)
      fprintf(out_fp, H2_P_HW_V);
    fprintf(out_fp, H2_TAG);

#if defined(GPFS_LINUX) && defined(LXTRACE_USE_RELAY)
    //for multiple raw trace file, read first trace record
    if (nCPU > 1)
    {
      cpuHdr[curCPU].hdr.trMagic = LXTRACE_SEQ_MAGIC;
      for (i=curCPU+1; i<nCPU; i++)
        {
          if (ReadNextTrcHeader(raw_fp[i], &cpuHdr[i], i, prevSeq)<0)
            cpuHdr[i].hdr.trMagic = LXTRACE_NODATA_MAGIC;
        }
    }
#endif

start_formatting:
    nWraps = 0; /* Reset the ReadNextTrcHeader wrap-found indicator. */

    /* Process raw trace records. Start by reading the device header. */
    while (0 < ( h = PickNextTrcHeader(raw_fp, nCPU, 
                                       cpuHdr, hasWrapped, prevSeq)))
      {
      //trace records should not be missed if there's only one trace file
      //the per-CPU trace file may have this problem due to trace disbalance on each CPU
      if( nCPU ==1 && h->hdr.trMagic == LXTRACE_SEQ_MAGIC)
      {
        if(h->hdr.trSeq - lastSeq > 1 && lastSeq != 0)
        {
          fprintf(out_fp,"*** Missing %llu trace records ***\n", h->hdr.trSeq - lastSeq - 1 );
          lostRecords += ( h->hdr.trSeq - lastSeq - 1 );
        } 
        lastSeq = h->hdr.trSeq;
      }

      recP = (trc_datahdr_t*) ( (char*)h + hdrSize[curCPU] - sizeof(trc_datahdr_t));
      DFP("rec.trHook 0x%X rec.trNArgs %d rec.trSPos %d rec.trSLen %d\n",
          recP->trHook, recP->trNArgs, recP->trSPos, recP->trSLen);

      if (h->hdr.trLength < sizeof(trc_datahdr_t))
      {
        DP("lxtrace format: bad record length in %s at offset %d, trLength %d\n",
           rawTraceFileNameP, ftell(raw_fp[curCPU]), h->hdr.trLength);
        rc = EINVAL;
        if (hasWrapped[curCPU])
        {
          DP("%s", "lxtrace format: restart from wrap-point\n");
          fprintf(out_fp, "*** BAD trace record header ***\n");
          goto do_wrap;
        }
        goto exit_format;
      }

      /* application_data[] holds the raw trace data that follows the headers. */
      int dataRead = 0;

      if (0 > (dataRead = fread(application_data, 1,
                          h->hdr.trLength - sizeof(trc_datahdr_t), raw_fp[curCPU])))
      {
        DP("lxtrace format: could not read data from %s at offset %d, rc %d errno %d\n",
           rawTraceFileNameP, ftell(raw_fp[curCPU]), dataRead, errno);
        rc = errno;
        goto exit_format;
      }

      DFP("read %d bytes of application trace data\n", dataRead);

      /* If this is the first trace record, consider this time zero,
         unless the -T option was specified */
      if (tzero.tv_sec == 0  &&  !useTOD)
        tzero = h->hdr.trTime;

      /* Convert trace time to time since first trace record unless
         the -T option was specified */
      if (!useTOD && !useLocal)
      {
        h->hdr.trTime.tv_sec -= tzero.tv_sec;
        if (h->hdr.trTime.tv_usec >= tzero.tv_usec)
          h->hdr.trTime.tv_usec -= tzero.tv_usec;
        else
        {
          h->hdr.trTime.tv_sec--;
          h->hdr.trTime.tv_usec += 1000000 - tzero.tv_usec;
        }
      }

      /* Get the formatting record for the trace record with this hookid */
      if (0 != GetFormatRecord(fmt_fp, (int)(recP->trHook & 0xFFFFFFFF),
                               &compP, &fmtP, &llmask, &i2nP, dataRead + hdrSize[curCPU]))
      {
        DP("lxtrace format: No formatting record found for hook %X\n",
           recP->trHook);
        fprintf(out_fp, "Unknown hook id 0x%08X\n", recP->trHook);
      }
      else
      {
        /* Found the formatting information for the current trace record. */

        /* Write leading information before the formatted trace. */
#ifdef GPFS_SOLARIS
        if (verbose)
          rc = fprintf(out_fp, fmtString,
                  h->hdr.trTime.tv_sec, h->hdr.trTime.tv_usec, h->hdr.trProcess,
                  h->hdr.trThread, h->hdr.trCPU, h->hdr.trBuf, recP->trHook,
                  compP);
        else
          rc = fprintf(out_fp, fmtString,
                  h->hdr.trTime.tv_sec, h->hdr.trTime.tv_usec, h->hdr.trProcess,
                  h->hdr.trThread, compP);
#else
        if (verbose)
        {
          if(sequence)
            rc = fprintf(out_fp, fmtString,
                  h->hdr.trSeq, h->hdr.trTime.tv_sec, h->hdr.trTime.tv_usec, h->hdr.trProcess,
                  h->hdr.trCPU, h->hdr.trBuf, recP->trHook,
                  compP);
          else
            rc = fprintf(out_fp, fmtString,
                  h->hdr.trTime.tv_sec, h->hdr.trTime.tv_usec, h->hdr.trProcess,
                  h->hdr.trCPU, h->hdr.trBuf, recP->trHook,
                  compP);
        }
        else
        {
          if(sequence)
            rc = fprintf(out_fp, fmtString,
                  h->hdr.trSeq, h->hdr.trTime.tv_sec, h->hdr.trTime.tv_usec, h->hdr.trProcess,
                  compP);
          else
          {
            if(useLocal)
              rc = fprintf(out_fp, fmtString,
                    fmtTime(h->hdr.trTime.tv_sec,timeBufP), h->hdr.trTime.tv_usec, h->hdr.trProcess,
                    compP);
            else
              rc = fprintf(out_fp, fmtString,
                    h->hdr.trTime.tv_sec, h->hdr.trTime.tv_usec, h->hdr.trProcess,
                    compP);
          }
        }
#endif
        if(rc<0)
        {
          DP("lxtrace format: fprintf error %s\n",strerror(errno));
          goto exit_format;
        }
        DFP("format string: %s\n", fmtP);
        DFP("raw data: %X %X %X %X %X\n",
            TRC_ARG(0), TRC_ARG(1), TRC_ARG(2), TRC_ARG(3), TRC_ARG(4));

        /* Format-specific processing. */
        if (recP->trSPos == _TR_FORMAT_I)
        {
          /* Trace data consists of n integer values */
          for (i=0; i < recP->trNArgs; i++)
            argP[i] = TRC_ARG(i);
        }
        else if (recP->trSPos < LXTRACE_MAX_FORMAT_SUBS)
        {
          /* Trace data contains null terminated string at TRC_ARG(rec.trSPos);
             the length of the string is given by rec.trSLen (rounded up to a
             multiple of ARGLEN).  Get a pointer to the string, make sure it's
             null-terminated (so we don't blow up if the trace data is trash),
             and compute how many extra TRC_ARG words are occupied by the
             string. */
          // ?? should validate trSLen!
          char *s = (char *)&TRC_ARG(recP->trSPos);
          int xi = recP->trSLen/ARGLEN - 1;
          s[recP->trSLen - 1] = '\0';

          for (i = 0; i < recP->trSPos; i++)
            argP[i] = TRC_ARG(i);
          argP[i++] = (argType)s;
          for (; i <= recP->trNArgs; i++)
            argP[i] = TRC_ARG(i + xi);
        }
        else if (recP->trSPos == _TR_FORMAT_F)
        {
          /* First n words are integer values; a double value is stored at
             TRC_ARG(rec.trNArgs) */
          double *d = (double *)&TRC_ARG(recP->trNArgs);

          for (i = 0; i < recP->trNArgs; i++)
            argP[i] = TRC_ARG(i);
          memcpy(&argP[i], d, sizeof(double));
        }
        else if (recP->trSPos == _TR_FORMAT_X)
        {
          /* XTrace type (preformatted) record. */
          fputs(application_data, out_fp);
          i = strlen(application_data);
          if (i == 0 || application_data[i-1] != '\n')
            putc('\n', out_fp);
          continue;
        }
        else
        {
          DP("lxtrace format: format %d not supported\n", recP->trSPos);
          fprintf(out_fp, "Unsupported format %d\n", recP->trSPos);
          goto exit_format;
        }

        /* Perform id2name substitutions */
        /* Defect 801461, always do this check instead of having it under DEBUGASSERTS*/
        if (i2nP == NULL &&
            (recP->trSPos==_TR_FORMAT_I || recP->trSPos==_TR_FORMAT_F) &&
            strstr(fmtP, "%s"))
        {
          DP("lxtrace format: hook %X is %s format record containing a string specification.\n",
             (recP->trHook&0xFFFFFFFF),
             (recP->trSPos==_TR_FORMAT_I) ? (char *) "an integer" :
               (recP->trSPos==_TR_FORMAT_F)? (char *) "a float":
               (char *) "an unknown");
          fprintf(out_fp, "lxtrace format: hook %X is %s format record with %d integer args containing a string specification.\n",
             (recP->trHook&0xFFFFFFFF),
             (recP->trSPos==_TR_FORMAT_I) ? (char *) "an integer" :
               (recP->trSPos==_TR_FORMAT_F)? (char *) "a float":
                  (char *) "an unknown",
                  recP->trNArgs);
          //print to both the outfile and stderr
          DP("%s", "               The formatting information for this record does not provide any substitutions.\n");
          fprintf(out_fp,"%s", "The formatting information for this record does not provide any substitutions:\n");
          fprintf(out_fp, "%s\n", fmtP);
          continue;
        }
        ID2Name(fmt_fp, recP, i2nP, id2nameP, argP);

#if defined(GPFS_LITTLE_ENDIAN) && !defined(__64BIT__) && !defined(_LP64)
        /* On a 32 bit machine, 64 bit values are stored with the high-order
           word first, followed by the low-order word.  The problem is that
           the reverse is expected by printf formatting.  For each "%ll" in
           the formatting string, swap the words in the argP array so that
           the words are swapped into the expected order.  The llmask
           bitvector indicates which arguments are of the %ll form. */
        if (llmask)
        {
          int tmp;
          int argPos = 0;

          DFP("\nlong long formatting: %s\n", fmtP);

          /* Loop through the bits in llmask */
          while (llmask)
          {
            if (llmask & 1)
            {
              DFP("switching argP[%d] and argP[%d]\n", argPos, argPos+1);
              tmp = argP[argPos];
              argP[argPos] = argP[argPos+1];
              argP[argPos+1]  = tmp;
              argPos += 1; /* account for the extra arg for this 2-word sub */
            }
            argPos += 1;
            llmask >>= 1;
          }
        }
#endif // GPFS_LITTLE_ENDIAN and NOT __64BIT__

        /* Output the completed trace record */
#if !defined(GPFS_ARCH_POWER) || !defined(GPFS_LINUX) || defined(__64BIT__)
        fprintf(out_fp, fmtP, TRC_ARGLIST);
#else
        /* On Linux 32-bit Power architecture (and maybe others?), the
           va_list calling convention does not allow passing a mixture of
           longs and long longs in successive memory words.  If long longs
           are present, break up the format string at each "%" and print
           the arguments separately, using llmask to determine which ones
           are long long. */
        if (llmask == 0)
          fprintf(out_fp, fmtP, TRC_ARGLIST);
        else
        {
          int n = 0, ll = llmask;
          unsigned long long bign;
          char *ptr, *newptr;
          char *newfmtP = malloc(strlen(fmtP) + 1);
          if (newfmtP == NULL)
          {
            DP("lxtrace format: unable to malloc format buffer");
            rc = ENOMEM;
            goto exit_format;
          }
          ptr = fmtP;
          newptr = newfmtP;
          while (*ptr)
          {
            if (*ptr != '%')
              *newptr++ = *ptr++;
            else
            {
              do
                *newptr++ = *ptr++;
              while (*ptr && *ptr != '%');
              *newptr = '\0';
              if (ll & 1)
              {
                bign = ((unsigned long long) argP[n] << 32) +
                  (unsigned long)argP[n+1];
                fprintf(out_fp, newfmtP, bign);
                n += 2;
              }
              else
              {
                fprintf(out_fp, newfmtP, argP[n]);
                n++;
              }
              ll >>= 1;
              newptr = newfmtP;
            }
          }
          *newptr = '\0';
          if (newptr != newfmtP)
            fprintf(out_fp, newfmtP);
          free(newfmtP);
        }
#endif
        nRecordsProcessed += 1;

      } /* end processing this trace record */
    } /* while there are more unformatted trace records to process */

    if (hasWrapped[curCPU])
    {
do_wrap:
      rewind(raw_fp[curCPU]); /* On a wrapped file, need to go back to the top. */
      hasWrapped[curCPU] = false;  /* Clear this so we don't do this more than once. */
      DFP("%s", "lxtrace format: done with wrap-point to EOF.  Going back to record 0\n");
      fileWrapped++;
      goto start_formatting;
    }

    rc = 0;

exit_format:
    
    /* Lost records count (primarily for per-cpu traces). The
       rest of the associated trace metadata is at the top of the file */
    if (out_fp != NULL)
      fprintf(out_fp, FMT_LOST_COUNT, lostRecords);

    /* print hash lookup statistics */
    DFP("Stats: hook lookup %6d check %7d (%.1f)\n",
        nHookLookup, nHookCheck, (double)nHookCheck/(double)nHookLookup);
    DFP("        i2n lookup %6d check %7d (%.1f) fail %d\n",
        nI2NLookup, nI2NCheck, (double)nI2NCheck/(double)nI2NLookup, nI2NFail);

    DFP("Wrap point: %d.%06d-%d.%06d (delta %d.%06d)\n",
        wrap_prev.tv_sec, wrap_prev.tv_usec, wrap_next.tv_sec, wrap_next.tv_usec,
        (wrap_prev.tv_sec-wrap_next.tv_sec), (wrap_prev.tv_usec-wrap_next.tv_usec));
    if (genStats)
    {
      int ctr;
      qsort(hookHashP, HOOK_HASH_TABLE_SIZE-1, sizeof(fmtDesc_t *), hookhashComparator);

      for (ctr = 0; ctr < HOOK_HASH_TABLE_SIZE-1; ctr++)
      {
        if (hookHashP[ctr] != NULL)
        {
          if (hookHashP[ctr]->numHits > 0)
          {
            fprintf(stat_fp, "hookid: 0x%X fmtP: %s     Hits: %d Total space: %lld bytes Average: %.2f Largest single entry: %d\n     Percent of uncompressed trcfile used: %.2f%% Percent of total traces: %.2f%%\n",
               hookHashP[ctr]->hookid,hookHashP[ctr]->fmtP,
               hookHashP[ctr]->numHits,hookHashP[ctr]->totalSpace,
               (double)hookHashP[ctr]->totalSpace/hookHashP[ctr]->numHits,hookHashP[ctr]->largestEntry,
               (double)hookHashP[ctr]->totalSpace/totalTrcfileProcessedSize*100,
               (double)hookHashP[ctr]->numHits/nRecordsProcessed*100);
          }
        }
      }
      fprintf(stat_fp, "Uncompressed trcfile space: %lld bytes. Compressed: %lld bytes. %lld traces seen\n", 
              totalTrcfileProcessedSize, totalCompSize,nRecordsProcessed);
      int elapsed = stopTime - startTime;
      if(elapsed == 0)
      {
        fprintf(stat_fp, "Elapsed time is less than 1 second. Assume it's 1 second for calculations\n");
        elapsed = 1;
      }
      fprintf(stat_fp, "Traces cover %d seconds. Trace throughput was %.1f traces/second\n",
         elapsed, (double)nRecordsProcessed/elapsed);
      fprintf(stat_fp, "Uncompressed data: %.1f KB/sec, compressed data: %.1f KB/sec\n",
         (double)totalTrcfileProcessedSize/elapsed/1000, (double)totalCompSize/elapsed/1000);
      fprintf(stat_fp, "Nota Bene:\n");
      fprintf(stat_fp, "  Each trace entry comes with a header of %d bytes\n",hdrSize[0]);
      fprintf(stat_fp, "  Hits ==> total number of times this hookid was seen\n");
      fprintf(stat_fp, "  Total space ==> total space in the trcfile used by all traces with this hookid\n");
      fprintf(stat_fp, "  Average ==> total space / hits\n");
    }
    /* cleanup */
    if (formatFileP) free(formatFileP);
    if (fmtDescSpaceP) free(fmtDescSpaceP);

    for (i=0; i<nCPU; i++)
      if (raw_fp[i] != NULL) fclose(raw_fp[i]);
    if (fmt_fp != NULL) fclose(fmt_fp);
    if (out_fp != NULL)
    {
      fflush(out_fp);
      fclose(out_fp);
    }
  } /* "format" option */


  /* Display the state information associated with the device. */
  else if (strcmp(argv[1], "dump") == 0)
  {
    int fd;
#define DUMP_LEN 512
    char dumpBufP[DUMP_LEN];

    fd = open(TRC_DEVICE, O_RDONLY);
    if (fd < 0)
    {
      DP("lxtrace dump: device open failed errno %d\n", errno);
      rc = ENODEV;
      goto exit;
    }

#if defined(GPFS_LINUX) && defined(LXTRACE_USE_RELAY)
    if ( RelayDump() < 0 )
#endif
    {

      /* Dump the trace device information */
      args.arg1 = DUMP_LEN;
#if defined(GPFS_LINUX)
      args.arg2 = (long)dumpBufP;
#elif defined(GPFS_SOLARIS)
      args.arg2 = (uint64_t)dumpBufP;
#endif
      if (0 != (rc = ioctl(fd, trc_dump, &args)))
      {
        DP("lxtrace dump: device dump returns rc %d errno %d \n", rc, errno);
      }
      else
      {
        printf("%s\n", dumpBufP);
      }
    }

exit_dump:
    if (fd >= 0)
      close(fd);
  }

  /* Flush any data that is currently buffered in the device. */
  else if (strcmp(argv[1], "fsync") == 0)
  {
    int fd;

    fd = open(TRC_DEVICE, O_RDONLY);
    if (fd < 0)
    {
      DP("lxtrace fsync: device open failed with errno %d\n", errno);
      rc = ENODEV;
      goto exit;
    }

#ifdef GPFS_SOLARIS
    if (0 != (rc = ioctl(fd, trc_fsynch, &args)))
#else
    if (0 != (rc = fsync(fd)))
#endif
    {
      DP("lxtrace fsync: device fsync returns rc %d errno %d\n", rc, errno);
    }

exit_fsync:
    if (fd >= 0)
      close(fd);
  }

  else if (strcmp(argv[1], "recycle") == 0 ||
           strcmp(argv[1], "off") == 0)
  {
    int fd;
    key_t key;
    int msgQueue;
    int retry = 0;
    trc_msg msg;
    struct stat buf;
    time_t mtime;

    msgQueue = msgGet(IPC_CREAT | 0666);
    if( msgQueue == -1 )
    {
      DP("lxtrace %s: message queue creation failed with errno %d\n", argv[1], errno);
      rc = errno;
      goto exit;
    }  
    DFP("lxtrace %s: message queue %d created\n",argv[1], msgQueue);    

    fd = open(TRC_DEVICE, O_RDONLY);
    if (fd < 0)
    {
      rc = ENODEV;
      goto exit_recycle;
    }

    
    if (strcmp(argv[1], "recycle") == 0 )
    {
      if (0 != (rc = ioctl(fd, trc_recycle, NULL)))
      {
        DP("lxtrace recycle: device recycle returns rc %d errno %d\n", rc, errno);
        goto exit_recycle;
      }
    }
    else
    { 
      /* trc_end terminates tracing. */
      args.arg1 = 0;
      rc = ioctl(fd, trc_end, &args);
      if (rc < 0)
      {
        DP("lxtrace off: device end failed with errno %d\n", errno);
        rc = EINVAL;
        goto exit_recycle;
      }
    } 

    if(argc>2 && stat(argv[2],&buf) == 0)
    {
      mtime = buf.st_mtime;
      DFP("lxtrace %s: trace file %s mtime %ld\n",argv[1],argv[2],buf.st_mtime); 
    }
    while(1)
    {   
      rc = msgrcv(msgQueue, &msg, 0, LXTRACE_MTYPE, IPC_NOWAIT); 
      if(rc == -1)
      {
        rc = errno;
        //no message, sleep for 1/10th second and retry
        if(errno == ENOMSG )
        {
          if(retry++ >= 200)
          {
            if(argc>2 && stat(argv[2],&buf) == 0)
            {
              DFP("lxtrace %s: trace file %s mtime %ld\n",argv[1],argv[2],buf.st_mtime); 
              //trace file is still being updated, wait for another round
              if(buf.st_mtime > mtime)
              {
                mtime = buf.st_mtime;
                retry = 0;
                continue;
              }
              else
                DFP("lxtrace %s: trace file %s not updated in last 20 seconds\n",argv[1],argv[2]); 
            }
            DFP("lxtrace %s: msgrcv timeout error %d\n",argv[1],errno);
            break;
          }
          else
            usleep(100000);
          DFP("lxtrace %s: msgrcv return %d\n",argv[1],errno);
        }
     
        else
        {
          DP("lxtrace %s: msgrcv error %d\n",argv[1],errno);
          break;
        }
      }
      //get desired message
      else 
      {
        DFP("lxtrace %s: msgrcv succeed\n",argv[1]);
        break;
      }
    }
exit_recycle:
    msgctl(msgQueue,IPC_RMID,NULL);
    close(fd);
  }
  else
    Usage();

exit:
  return rc;
}
#endif //GPFS_WINDOWS
